.include	"m1284Pdef.inc"
;--------------------------------------------------------------------------
;
;	ROMulus the 1st
;
;	ROMulus is a W65C816 based single board computer. The main features
;	are:
;		-	ROM less design, a ROM image is downloaded via IML / blind
;			load to RAM which is later write protected
;		-	PS/2 keyboard interface
;		-	24x40 and 24x80 character VGA display
;
;	These features are implemented using a ATmega1284P microcontroller.
;	The microcontroller is clocked using a 22.1184MHz baud-rate XTAL
;	and controls the download of ROM images, video signal generation and
;	the handling of the PS/2 interface.
;
;	Currently the ATmega1248P also controls PHI2. During the IML it
;	controls the state of PHI2 using the normal function of IO pins.
;	The pin used is the one that has OC2A as alternate function. Once
;	the ROM has been loaded you can activate OC2A to act as PHI2. The 
;	clock of PHI2 can be set as high as half the clock of the MCU.
;
;	At the moment you need to connect a serial terminal to the USART0 of
;	the ATmega1284P to control and boot the SBC.
;
;--------------------------------------------------------------------------
;
;	To do list
;
;	-	Initial commmand, as in the prototype of AVRIML an initial command
;		should be configurable so that the system power-ups with the
;		default ROM without need to connect a serial terminal
;
;	-	Add the video control signal buffer and add software support for 
;		the various modes.
;
;	-	A nice boot and setup menu that uses the VGA display feature and 
;		the PS/2 key board prior to starting the SBC. Alternatively the 
;		same boot menu should be accessible via USART0.
;
;	-	Support for linear frame buffers. Currently video RAM is addressed
;		as in a Apple IIe. The reason for this is that the AVR can only
;		update 8-bits of the video RAM for one chracter during the display
;		phase. That means all characters of a row must be within a block
;		of 256bytes. 
;
;	-	Add support for external oscillator to have higher PHI2 clock 
;		rates, currently the maximum clock rate is half the clock rate
;		of the AVR
;
;	-	Add support for low-res graphics like in a Apple II/IIe.
;
;--------------------------------------------------------------------------
;
;	Change history
;
;
;	2015-09-20	Stop the timers and not only stop interrupts as after it
;				is reactivated it may fire directly with a counter value
;				that results in a invalid de-jitter offset.
;
;	2015-09-19	Use GPIOR0 for KWAIT and GPIOR1 for KBUF to save cycles
;
;	2015-09-17	New logic to generate the timer based VGA signals so you
;				can dynamically switch the VGA signal generation off and
;				on again. This also allows to dynamically change the
;				parameters.
;
;	2015-09-15	New project as an offspring from vgacontrollerIV now called
;				ROMulus 1st
;
;	2015-09-10	Use Timer 3 for SHLD and not Timer 2, the reason is that
;				in fast PWM mode OCR3A cannot be set lower than 0x0003
;				which would create a PHI2 of only 1/4 of the AVR clock
;				speed. Timer2 is better suited for this purpose and can
;				create a PHI2 of up to 1/2 of the AVR clock. 
;
;	2015-09-08	Add again a 74HCT573 Output PIN extension referred to as
;				PORTE. PD7 is used as latch enable and the inputs are
;				connected to PORTA, which is always configured as outputs.
;				this frees up 6 pins form PORTB. These pins are now used
;				to implement the IML/blind-load feature of the AVRIML
;				project. 
;
;	2015-09-02	New project as an offspring from old vgacontrollerII now
;				called vgacontrollerIV. VGA controller II had a dedicated
;				port that connected to the 74HC166. VGA controller IV
;				now uses PORTA to time mutliplex the lower 8 bit video
;				RAM address and the character bit map data sent to the
;				74HC166 parallel to serial shift register. Therefore the
;				address latch used in VGA controller II is no longer
;				required.
;
;--------------------------------------------------------------------------
;
;	VGA Controller
;
;	The VGA controller uses a dual-port RAM (in our case it is a 4kbyte
;	IDT7134) to form a frame-buffer. The IDT7134 has no arbitration logic
;	and therefore is best suited when there is no need to handle access
;	conflicts. So far I have never seen artifacts in the VGA display even
;	when the SBC scrolls the data.
;
;	The video memory is mapped like in a Apple II. The controller supports
;	the following features of Apple II compatible display
;	-	40 characters line and 24 lines b/w text display
;	-	support for PAGE1 and PAGE2
;	-	when A11 of the SBC side of the dual-port RAM is connected to
;		the main/aux selection the controller supports 80 characters
;		per line with a layout identical to the Apple IIe video RAM
;		mapping
;
;	We use a 74HC166 to convert the parallel video data to a serial
;	data stream. The timing is handled using three timers.
;
;	Timer1	This timer is used in fast-PWM mode with ICR1A set as the
;			top value. OC1A is used to create HSYNC using OCR1A you can
;			adjust the width of the HSYNC pulse. OCR1B triggers the       
;			interrupt that performs the various task to generate the 
;			video signal. 
;
;	Timer3	To convert the bitmap form the CGROM to a pixel data stream
;			I use a 74HC166. This is a fully synchronous paralle to
;			serial shift register. That means not only serial shift but
;			as well parallel load is using the clock. To switch between
;			shift and load you have to apply the appropriate level to
;			SHLD. The 74HC166 is clocked with the ATmega1248P clock, 
;			which is also the pixel clock and we have 7 pixels per
;			character. The SHLD signal must in this case be low for
;			the first clock to load the data and then it must be high
;			for the subsequent 6 clocks. We use Timer 3 to create such
;			a signal. Timer 3 is configured in fast-PWM mode with OCR3A
;			set as TOP (in our case this is 6 = pxlpch-1, pixel per
;			character). SHLD is generated using OC3B and OCR3B set to 
;			one less than OCR3A. OC3B is clreared on a match and set
;			at bottom.
;
;	Timer0	The XTAL oscillator of the ATmega1284P is configured to
;			run in full swing mode so we can use XTAL2 as clock for
;			the 74HC166. To support a 40-column we use the INH clock
;			input of the 74HC166. INH is set high during every second
;			clock. For this we use Timer0. We run the timer in fast-PWM
;			mode with OCR0A set to 1. OC0B is connected to INH
;			In 80-column mode OC0B is disconnected from the timer and
;			permanently set low, so the pixel clock is equal to the
;			crystal clock. 
;			In 40-column mode OC0B is connected to the timer and toggles
;			with every clock cycle of the ATmega. Thus it inhibits every
;			second clock.
;
;	The controller creates a 480*640@60Hz compatible VGA signal. That is
;	it uses the standard timing for HSYNC and VSYNC. But it only creates
;	560pixels per scan line. This is because I wanted to use CGROMs of
;	legacy systems that only use 280 or 560 pixels for their video output
;	and allows me to get the same patterns when using special characters
;	for semi-graphic output.
;
;--------------------------------------------------------------------------
;
;	ATmega1248P Port Usage
;
;	PORTA	This port is used as a 8-bit output port. It is permanently
;			configured as ouput. It is connected to the paralle linputs
;			of the 74HC166, to 8 address lines of the dual-port RAM,
;			to the inputs of the 74HC574 IML and PS/2 keyboard latch
;			and to the inputs of a 74HC573 ouput port expansion latch 
;			called PORTE
;
;			During horizontal display the controller time multiplexes
;			the lower 8 bit if the video RAM address and the CGROM bitmap
;			data. For this to work and to achieve a jitter free display
;			the time multiplexing is 100% synchronized to Timer 3.
;
;			When the display is blanked the microcontroller uses PORTA
;			to control the 74HC573 output port expansion. This port
;			expansion is called PORTE and contains the remaining 4
;			video RAM address bits, /OE and RW of the dual-port RAM
;			and VSYNC.
;	
;			The 74HC574 which is also connected to PORTA serves two
;			functions. During the load of the ROM image it serves as
;			the data buffer and in normal operation it is used to
;			store the ASCII value of the key pressed on the PS/2 
;			keyboard. Vidoe is switched off when the latch is used
;			to load the ROM image. When used as a keyboard buffer the
;			video signal is activated as this features dos not affect
;			the timing of the video signal.
;
;	PORTB	various signal see description at initialisation of PORTB.
;
;	PORTC	This port is connected to the data signals of the dual-port
;			RAM and also to the outputs of a video control signal
;			latch. The video control signal latch is used to control
;			the various display modes. Normally PORTC is configured
;			as input and we use CE and INP to control the ouput of
;			the dual-port RAM and the video control signal buffer. 
;			However it is possible to write to the dual-port RAM as
;			RW is connected to the port expansion PORTE. The video 
;			control signals are referred to as PORTF.
;
;	PORTD	various signal see description at initialisation of PORTD.
;
;--------------------------------------------------------------------------
;
;--------------------------------------------------------------------------
;
;	Register Definitions
;
;	Some registers are exclusively used by certain parts of the program.
;	They are never saved or restored and must only be used by the part
;	of the code that deals with the feature. We also make use of many
;	registers as constants to save cycles and to make the code more
;	readable.
;
.def	lemask	= r4			; constant (1<<LE)
.def	zero	= r5			; constant 0x00
.def	one		= r6			; constant 0x01
.def	ff		= r7			; constant 0xff
;
;		PORTE and PORTF are used globally. They are internally copies
;		of the output port expansion and the video control signals
;
.def	PORTE	= r9			; Output Port Expansion
.def	PORTF	= r10			; Video Control Input Flags
;
;		'flash' is incremented after one screen has been displayed, that
;		is every 16ms. Bit 5, which changes approximatively every 500ms
;		is used to switch between the normal and inverse CGROM to produce
;		flashing characters. This register is only used by the horizontal
;		interrupt routine.
;
.def	flash	= r12			; 
;
;		These are as the flags from the keyboard processing. These
;		registers are only used by the PS/2 processing routines.
;
.def	Flagnormal	= r13		; Normal Keyboard Flags (Shift, Ctlr, ...)
.def	Flagtoggle	= r14		; Toggled Keyboard Flags (Numlock, Capslock, ...)
;
;		This temporary register is only used in normal code
;
.def	temp	= r16			; Temp, w/ immediate support
;
;		The following registers are only used by the horizontal interrupt
;		routine (except for line which is initialised together with the 
;		video state machine)
;
.def	line 	= r17			; general purpose line counter
.def	scan 	= r18			; scan line counter for character display
.def	vd		= r19			; video data
.def	itemp	= r20			; Temp w/ immediate support
;
;		CHAR is used as input/output parameter for the serin and serout
;		routines.
;
.def	char	= r21			; USART character
;
;		The following registers are only used by the PS/2 processing
;		routines
;
.def	Scancode	= r22		; Scancode
.def	Counter		= r23		; Loop Counter
;
;	GPIOR0 definitions
;
.equ	KWAIT		= 0
.equ	JITTER		= 1

;--------------------------------------------------------------------------
;
;	Macro Definitions
;
.macro	isave			;		Interrupt save shared register
	push	r0			;  2
	in		r0, SREG	;  1
	push	xl			;  2
	push	xh			;  2
	push	zl			;  2
	push	zh			;  2
.endmacro				; 11
;
;
;
.macro	irestore		;		Interrupt restore shared register
	out		SREG, r0	;  1
	pop		zh			;  2
	pop		zl			;  2
	pop		xh			;  2
	pop		xl			;  2
	pop		r0			;  2
.endmacro				; 11
;
;	Align to next address which is a multiple of 2**N (N=1...15) 
;
.macro	align             ;align to (1<<@0)
alignfromhere:
      .if (alignfromhere & ((1<<@0)-1))   ;if not already aligned
         .org  (alignfromhere & (0xffff<<@0)) + (1<<@0)
      .endif
.endmacro
;
.macro	scanto
;
;	Convert VGA 'scan' line to character ROM scan line offset. The
;	typical eight scan lines per character in a legacy CGROM must
;	be mapped to the 20 VGA scan lines.
;
	ldi		zl, low(2*scantotable)	;  1
	ldi		zh, high(2*scantotable)	;  1
	add		zl, scan				;  1
	adc		zh, zero				;  1
	lpm		@0, Z					;  3
.endmacro							;  7	cycles
;
;	Convert Character 'line' number to 16-bit address in X
;
.macro	linetox
;
;	Convert Line (line) number to 16-bit Video Address (X) to achieve
;	the same mapping as is used by the Apple II Monitor Routine BASCALC
;
;	000ABCDE	->	000000CD EABAB000
;
;	However in our case A7,A8,A9 (EDC) are connected to the lower three
;	bits of the	high-byte address extension and A11 of the DP-RAM is
;	connected to PA7. Therefore we need to do the following conversion.
;
;	000ABCDE	->	00000CDE 0ABAB000
;
	mov		xl, line			;  1
	andi	xl, 0x18			;  1	isolate AB
	mov		xh, xl				;  1	000AB000
	lsl		xl					;  1
	lsl		xl					;  1	0AB00000
	or		xl, xh				;  1	0ABAB000
	mov		xh, line			;  1
	andi	xh, 0x07			;  1	00000CDE
.endmacro						;  8	cycles
;--------------------------------------------------------------------------
;
;	VSYNC:	The standard for 640x480@25MHz is 11/2/32, that is 45
;			non-display lines to make it a total of 525 lines as defined
;			in the VESA standard. These 45 non-display lines are:
;
	.equ	vfplines	= 11	; Front Porch Lines	
	.equ	vsynclines	= 2		; VSync Lines
	.equ	vbplines	= 32	; Back Porch Lines
;--------------------------------------------------------------------------
;
;	RAM
;
	.dseg
	.org		SRAM_START			; Make sure we start at SRAM
nextstate:		.byte	2			; Next State in VGA display
nexthstate:		.byte	2			; Next Hline State
tmrskew:		.byte	2			; Skew between Timer 1 and Timer 2
keyavr:			.byte	1			; Parallel print scancodes
t0info:			.byte	1
t1info:			.byte	2
t3info:			.byte	2
jtcnt1:			.byte	2
;
;	During the tests it was often required to ajdust values used for the
;	VGA timing. To avoid iterations with assembling and loading I just
;	define same variables in RAM that are initialised with some best guess
;	constants and could then be altered via internal monitor commands.
;
hline:			.dw		700			; Cycles per hline
hsyncw:			.dw		87			; Width of HSYNC pulse
dstart:			.dw		56			; Trigger point of hline ISR
shldoff:		.db		0			; Skew of SHLD to HSYNC
inhpol:			.db		0			; Polarity of INH
inhcon:			.db		0			; 0=disconnect (1<<COM0B1)=connect
pxlpch:			.db		7			; Pixels per character, 7 or  8
;
;--------------------------------------------------------------------------
;
;	Vectors, unused vectors, spurios interrupts result in a RESET
;
	.cseg
	.org	0x0000
	jmp		reset		;
	jmp		reset		; INT0addr	= 0x0002	; External Interrupt Request 0
	jmp		reset		; INT1addr	= 0x0004	; External Interrupt Request 1
	jmp		reset		; INT2addr	= 0x0006	; External Interrupt Request 2
	jmp		reset		; PCI0addr	= 0x0008	; Pin Change Interrupt Request 0
	jmp		reset		; PCI1addr	= 0x000a	; Pin Change Interrupt Request 1
	jmp		reset		; PCI2addr	= 0x000c	; Pin Change Interrupt Request 2
	jmp		reset		; PCI3addr	= 0x000e	; Pin Change Interrupt Request 3
	jmp		reset		; WDTaddr	= 0x0010	; Watchdog Time-out Interrupt
	jmp		reset		; OC2Aaddr	= 0x0012	; Timer/Counter2 Compare Match A
	jmp		reset		; OC2Baddr	= 0x0014	; Timer/Counter2 Compare Match B
	jmp		reset		; OVF2addr	= 0x0016	; Timer/Counter2 Overflow
	jmp		reset		; ICP1addr	= 0x0018	; Timer/Counter1 Capture Event
	jmp		reset		; OC1Aaddr	= 0x001a	; Timer/Counter1 Compare Match A
	jmp		ISR_hline	; OC1Baddr	= 0x001c	; Timer/Counter1 Compare Match B
	jmp		reset		; OVF1addr	= 0x001e	; Timer/Counter1 Overflow
	jmp		reset		; OC0Aaddr	= 0x0020	; Timer/Counter0 Compare Match A
	jmp		reset		; OC0Baddr	= 0x0022	; Timer/Counter0 Compare Match B
	jmp		reset		; OVF0addr	= 0x0024	; Timer/Counter0 Overflow
	jmp		reset		; SPIaddr	= 0x0026	; SPI Serial Transfer Complete
	jmp		reset		; URXC0addr	= 0x0028	; USART0, Rx Complete
	jmp		reset		; UDRE0addr	= 0x002a	; USART0 Data register Empty
	jmp		reset		; UTXC0addr	= 0x002c	; USART0, Tx Complete
	jmp		reset		; ACIaddr	= 0x002e	; Analog Comparator
	jmp		reset		; ADCCaddr	= 0x0030	; ADC Conversion Complete
	jmp		reset		; ERDYaddr	= 0x0032	; EEPROM Ready
	jmp		reset		; TWIaddr	= 0x0034	; 2-wire Serial Interface
	jmp		reset		; SPMRaddr	= 0x0036	; Store Program Memory Read
	jmp		reset		; URXC1addr	= 0x0038	; USART1 RX complete
	jmp		reset		; UDRE1addr	= 0x003a	; USART1 Data Register Empty
	jmp		reset		; UTXC1addr	= 0x003c	; USART1 TX complete
	jmp		reset		; ICP3addr	= 0x003e	; Timer/Counter3 Capture Event
	jmp		reset		; OC3Aaddr	= 0x0040	; Timer/Counter3 Compare Match A
	jmp		reset		; OC3Baddr	= 0x0042	; Timer/Counter3 Compare Match B
	jmp		reset		; OVF3addr	= 0x0044	; Timer/Counter3 Overflow
;
	.org	INT_VECTORS_SIZE		; Set origin after interrupt vector table
;--------------------------------------------------------------------------
;
;	Startup/Help screen
;
hello:	.db	0x0d, 0x0a, 0x0d, 0x0a
;			"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
		.db	"ROMulus 1st Version 0.7                                                         ", 0x0d, 0x0a
		.db	"                                                                                ", 0x0d, 0x0a
		.db	"     For general syntax see Apple II monitor. The following commands of the     ", 0x0d, 0x0a
		.db	"     Apple II Monitor are implemented: blank, dot, colon, -, +, <, move, verify ", 0x0d, 0x0a
		.db	"                                                                                ", 0x0d, 0x0a
		.db	"     The following commands are available to adjust VGA dispaly                 ", 0x0d, 0x0a
		.db	"     ?     this screeen                                                         ", 0x0d, 0x0a
		.db	"     ^T    Create Test Pattern 0x00..0xff at 0x1000                             ", 0x0d, 0x0a
		.db	"     n^O   shldoff                                                              ", 0x0d, 0x0a
		.db	"     n^P   inhpolarity      0=even, 1=odd                                       ", 0x0d, 0x0a
		.db	"     n^L   hline                                                                ", 0x0d, 0x0a
		.db	"     n^R   dstart                                                               ", 0x0d, 0x0a
		.db	"     n^Z   Set 40/80        0=40-column, 1=80-column                            ", 0x0d, 0x0a
		.db	"     ^A    Altchar (toggle)                                                     ", 0x0d, 0x0a
		.db	"     ^I    Info                                                                 ", 0x0d, 0x0a
		.db	"     n^N   print scancodes  0=no, 1=yes                                         ", 0x0d, 0x0a
		.db	"     n^V   Video on/off     0=off, 1=on                                         ", 0x0d, 0x0a
		.db	"                                                                                ", 0x0d, 0x0a
;		.dw	0
		.db	"     The following commands are available to control the 65xxx SBC              ", 0x0d, 0x0a
		.db	"     ^E    Erase lower memory for cold start                                    ", 0x0d, 0x0a
		.db	"     nR    Reset Control    0=de-assert (release), 1=assert (pull)              ", 0x0d, 0x0a
		.db	"     nY    Load Image       0=Apple II, 1=Apple IIplus                          ", 0x0d, 0x0a
		.db	"     I     Toglle IML                                                           ", 0x0d, 0x0a
		.db	"     nK    Set Clock        0=11MHz, 1=5.5MHz, a=1MHz, etc                      ", 0x0d, 0x0a
		.db	"     T     Stop Clock                                                           ", 0x0d, 0x0a
		.db	"     nW    65xxx Cycle with nn as data on the databus                           ", 0x0d, 0x0a
		.db	"     nO    set keyboard register value                                          ", 0x0d, 0x0a
		.db	"     P     Toggle PHI2                                                          ", 0x0d, 0x0a
		.db	"     S     Show Status of VPB, IML, PHI2 and RESET                              ", 0x0d, 0x0a
		.db	"     X     Download a ROM image via XModem and restart SBC                      ", 0x0d, 0x0a
		.db	"                                                                                ", 0x0d, 0x0a
		.dw	0
;--------------------------------------------------------------------------
;
;	The following table is used to convert the VGA scan line (0..19) to a
;	CGROM section offset (0..7)
;
scantotable:
	.db	0x00, 0x00, 0x01, 0x01, 0x01, 0x02, 0x02, 0x03, 0x03, 0x03
	.db	0x04, 0x04, 0x05, 0x05, 0x05, 0x06, 0x06, 0x07, 0x07, 0x07
;--------------------------------------------------------------------------
; Baud Rate Calculation
;--------------------------------------------------------------------------
;
;	The VGA controller uses a baud-rate crystal of 22.1184MHz
;
.equ	F_CPU	= 22118400		; CPU Clock Hz
.equ	BAUD	= 115200		; Baudrate of UART
;
; Baud Rate Calculator
;
.equ	UBRR_VAL	= ((F_CPU+BAUD*8)/(BAUD*16)-1)	; nearest value
.equ	BAUD_REAL	= (F_CPU/(16*(UBRR_VAL+1)))		; effective baudrate
.equ	BAUD_ERROR	= ((BAUD_REAL*1000)/BAUD-1000)	; error in 0/00
.if ((BAUD_ERROR>10) || (BAUD_ERROR<-10))
	.error " Effective baudrate is more then 1 percent off requested baudrate"
.endif
;--------------------------------------------------------------------------
;
;	Initialise the timers that generate the video display. We now use
;	three timers, one timer is responsible for the HSYNC, one timer is
;	responsible for the SHLD signal and one timer is responsible for the
;	INH signal. 
;
;	Different to previous approaches we now provide one subroutine that
;	initialises the timers. Also we no longer adjust the parameters 
;	dynamically but set the phases of the timers in this subroutine
;	according to the parameters given in a global section
;
.equ	i_hline		= 700			; hline cycles
.equ	i_hsyncw	= 700 - 83		; width of HSYNC
.equ	i_dstart	= 700 - 46		; start of OCR1B interrupt
.equ	i_shldoff	=   0			; skew between T3 and T1
.equ	i_inhpol	=   0			; polarity of T0 in respect to T1
.equ	i_pxlpch	=   7			; Number of Pixels per characgter
;
;	Initialise the timer values in RAM using the above constants
;
videoinit:
	ldi		temp, (1<<COM0B1)
	sts		inhcon, temp

	ldi		zl, low(hline_40c)
	ldi		zh, high(hline_40c)
	sts		nexthstate, zl
	sts		nexthstate+1, zh

	ldi		temp, low(i_hline)
	sts		hline, temp
	ldi		temp, high(i_hline)
	sts		hline+1, temp
	
	ldi		temp, low(i_hsyncw)
	sts		hsyncw, temp
	ldi		temp, high(i_hsyncw)
	sts		hsyncw+1, temp

	ldi		temp, low(i_dstart)
	sts		dstart, temp
	ldi		temp, high(i_dstart)
	sts		dstart+1, temp
	
	ldi		temp, i_shldoff
	sts		shldoff, temp
	
	ldi		temp, i_inhpol
	sts		inhpol, temp
	
	ldi		temp, i_pxlpch
	sts		pxlpch, temp
									; Fall through to videostop to 
									; make sure counters are all cleared
;--------------------------------------------------------------------------
;
;	Stop all timers and stop interrupt and clear the count register
;	When setting 16-bit timer registers you must write the high byte first
;	as the timer register is only updated when the low byte is written,
;	the high byte is taken by the previous high byte which was only 
;	written to the temp register internal to the AVR. This guarantees
;	that updates to 16-bit registers are atomic.
;
videostop:
	sts		TIMSK1, zero			; Disable hline ISR

	out		TCCR0B, zero
	out		TCCR0A, zero			; Stop Timer 0
	out		TCNT0, zero				; Set Count to 0

	sts		TCCR1B, zero
	sts		TCCR1A, zero			; Stop Timer 1
	sts		TCNT1H, zero			; Set Count to 0
	sts		TCNT1L, zero

	sts		TCCR3B, zero
	sts		TCCR3A, zero			; Stop Timer 3
	sts		TCNT3H, zero			; Set Count to 0
	sts		TCNT3L, zero

	ret
;
;--------------------------------------------------------------------------
;
;	Start all timers and interrupts required to generate the video signal.
;
;
;	INH		Timer 0
;	HSYNC	Timer 1
;	SHLD	Timer 3
;
;
; Timer 1
;		---------------------+    HSYNC      +-----------------------
; TCNT1	                     |               |
; 		                     +---------------+
;		                     ^      ^        ^
; 		                     hline  dstart   TOP   
;
; Timer 3
;		------+ +------+ +------+ +------+ +------+ +------+ +------+ +-
; SHLD	      | |      | |      | |      | |      | |      | |      | |
; 		      +-+      +-+      +-+      +-+      +-+      +-+      +-+
;             ^ ^
;      pxlpch-2
;        pxlpch-1
;
videostart:
	out		PORTA, PORTE
	sbi		PORTA, VSYNC			; VSYNC = high
	sbi		PORTD, LE
	cbi		PORTD, LE
	in		PORTE, PORTA
	ldi		line, 1
	ldi		zl, low(vbporch_d)
	ldi		zh, high(vbporch_d)		
	sts		nextstate, zl
	sts		nextstate+1, zh			; Initial state of video generator
;--------------------------------------------------------------------------
;
;	Timer0:	Generate INH, note that for the 80 col display OC0B is 
;			disconnected and set low and only for 40 col display OC0B
;			is connected. This will inhibit every second clock of the
;			74HC166, therefore we eventually need to adjust the phase
;			of Timer 0 to Timer 1 by one cycle (see below).
;
	out		OCR0B, zero
	out		OCR0A, one
	lds		temp, inhcon
	ori		temp, (0<<COM0A1) | (0<<COM0A0) | (0<<COM0B1) | (0<<COM0B0) | (1<<WGM01) | (1<<WGM00)
	out		TCCR0A, temp
	ldi		temp, (1<<WGM02) | (0<<CS02) | (0<<CS01) | (1<<CS00)
	out		TCCR0B, temp
;--------------------------------------------------------------------------
;
;	Optionally add one cycle before starting Timer1 so the phase of
;	Timer0 to Timer1 can be odd or even. 
;	
	lds		temp, inhpol
	sbrc	temp, 0			; skip takes 2 cycles when bit0 is cleared
	rjmp	nextcycle		; 2 cycle NOP
nextcycle:
;--------------------------------------------------------------------------
;
;	Timer1:	ICR1 is used as TOP, here we use a 22MHz crystal and
;	only 7 pixels per character. Note in order to be able to write
;	ICR1 you must first activate a wave-form that uses ICR1 as TOP!
;
	ldi		temp, (1<<WGM11)
	sts		TCCR1A, temp
	ldi		temp, (1<<WGM13) | (1<<WGM12)
	sts		TCCR1B, temp
	lds		zl, hline				; 
	lds		zh, hline+1				; 
	sbiw	zl:zh, 1				; Top must be set to one less than
	sts		ICR1H, zh				; the cycle count of the timer
	sts		ICR1L, zl				; 
;
;	Timer1: OC1A is used as HSYNC
;
	lds		temp, hsyncw+1
	sts		OCR1AH, temp
	lds		temp, hsyncw
	sts		OCR1AL, temp
;
;	Timer1: OCR1B is used to generate a regular interrupt
;
	lds		temp, dstart+1
	sts		OCR1BH, temp
	lds		temp, dstart
	sts		OCR1BL, temp
;
;	Enable OCB1 Match Interrupt 
;
	ldi		temp, 1<<OCIE1B
	sts		TIMSK1, temp
;	
;	Timer1: fast-PWM, prescaler = 1, TOP=ICRA, OC1A to create HSync
;			OC1A is used in non-inverting mode now, that is it is
;			cleared on compare match and set at BOTTOM
;
	ldi		temp, (1<<COM1A1) | (0<<COM1A0) | (1<<WGM11)
	sts		TCCR1A, temp
	ldi		temp, (1<<WGM13) | (1<<WGM12) | (1<<CS10)
	sts		TCCR1B, temp
;--------------------------------------------------------------------------
;
;	Wait a variable time of cycles before initialising Timer 3. This is
;	important as PORTA has only for one cycle the valid video data. This
;	valid cycle must be in sync to SHLD. Note that when you make changes
;	to the interrupt routines that create the display you must eventually
;	addjust shldoff. 
;
	lds		zl, shldoff
	andi	zl, 0x0f		; Just to make sure nothing can go wrong
	clr		zh
	subi	zl, low(-shldoffset)
	sbci	zh, high(-shldoffset)
	ijmp
shldoffset:
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
;--------------------------------------------------------------------------
;
;	Timer3:	Generate Load/Shift Signal on OC3B, 
;
	lds		temp, pxlpch
	dec		temp				; pxlpch-1
	sts		OCR3AH, zero
	sts		OCR3AL, temp
;
;	Timer3: OCR3B is used to generate SHLD
;
	dec		temp				; pxlpch-2
	sts		OCR3BH, zero
	sts		OCR3BL, temp
;	
;	Timer3: fast-PWM, prescaler = 1, TOP=OCRA, OC1B to create HSync
;
	ldi		temp, (1<<COM3B1) | (0<<COM3B0) | (1<<WGM31) | (1<<WGM30)
	sts		TCCR3A, temp
	ldi		temp, (1<<WGM33) | (1<<WGM32) | (1<<CS30)
	sts		TCCR3B, temp		; Start Timer 
;--------------------------------------------------------------------------
;
;	Get a snapshot before we enter just for tests purpose
;
	in		temp, TCNT0
	lds		r0, TCNT1L
	lds		r1, TCNT1H
	lds		r2, TCNT3L
	lds		r3, TCNT3H
	
	sts		t0info, temp
	sts		t1info, r0
	sts		t1info+1, r1
	sts		t3info, r2
	sts		t3info+1, r3
;--------------------------------------------------------------------------
	ret
;--------------------------------------------------------------------------
;
;	Display Info and Help screen on startup and on request.
;
splashscreen:
	ldi		zl, low(2*hello)
	ldi		zh, high(2*hello)
hloop:
	lpm		char, Z+
	tst		char
	breq	hdone			; 
	rcall	serout
	rjmp	hloop
;
hdone:						; 
	ret
;--------------------------------------------------------------------------
; 
;	Polled USART Output
;
;	A small note why we only use polled routines. The way I generated the
;	VGA signal does not allow other interrupts and does not even allow to
;	disable and re-enable interrupts. Else the de-jitter code in the ISR
;	(interrupt service routine9 will miserably fail!
;
;	We have in fact a main program which does the routine tasks that must
;	be interuptable at every instruction and the interrupt service routine
;	for the horizontal line processing. If we need the full attention of
;	the processor (e.g. during XModem download) we need to stop video
;	signal generaten and the restart it again.
;
seroutcrlf:
	ldi		char, 0x0a
	rcall	serout
	ldi		char, 0x0d
serout:
	push	char
seroutwait:
	lds		char, UCSR0A
	sbrs	char, UDRE0
	rjmp	seroutwait
	pop		char
	sts		UDR0,char
	ret			
;--------------------------------------------------------------------------
;
;	Wait for a character received via either UART0 (serial input) or 
;	or USART1 (PS/2 input). 
;
;	In case we have a receive error on USART1 (PS/2), print an error
;	message on USRAT0. It seems that I still have some issues with the
;	signal quality of PS/2 and this creates quite a lot of input errors!
;
serin1error:					; We got a PS/2 receive error
	push	temp
	mov		temp, char
	ldi		char, 'E'
	rcall	serout
	ldi		char, 'R'			;
	rcall	serout
	rcall	serout
	ldi		char, ':'			; Print "ERR:"
	rcall	serout
	ldi		char, '-'			; Assume no error
	sbrc	temp, DOR1
	ldi		char, 'O'			; 'O' for Data Overrun
	rcall	serout

	ldi		char, '-'
	sbrc	temp, FE1
	ldi		char, 'F'			; 'F' for Framing Error
	rcall	serout

	ldi		char, '-'
	sbrc	temp, UPE1
	ldi		char, 'P'			; 'P' for Parity Error
	rcall	serout

	rcall	seroutcrlf			; 
	lds		char, UDR1			; Reset USART status
	pop		temp
serin:
	lds		char, UCSR1A		; PS/2 input status
	sbrs	char, RXC1			; Scancode ready?
	rjmp	serin010			; No, test for RS-232

	andi	char, (1<<DOR1) | (1<<FE1) | (1<<UPE1)
	brne	serin1error			; we have error flags, not good

	push	r0					; Prepare to call state machine
	push	r1
	push	temp
	push	zl
	push	zh
	lds		Scancode, UDR1		; get scancode
	lds		r0, keyavr
	tst		r0
	breq	serin005
	mov		r0, Scancode
	rcall	prbyte
	rcall	seroutcrlf
serin005:
	lds		zl, ps2st
	lds		zh, ps2st+1			; execute next state
	icall
	pop		zh
	pop		zl
	pop		temp
	pop		r1
	pop		r0

serin010:
	lds		char, UCSR0A
	sbrs	char, RXC0
	rjmp	serin				; repeat until we have an input
	lds		char, UDR0
	ret
;--------------------------------------------------------------------------
;
;	After a hardware reset
;
reset:
;
;	Set Stack Pointer to top of RAM
;
	ldi		temp,low(RAMEND) 
	out		SPL,temp 
	ldi		temp,high(RAMEND) 
	out		SPH,temp
;
;	Init Constants
;
	clr		zero
	clr		flash
	clr		one
	inc		one
	clr		ff
  	dec		ff
  	ldi		temp, (1<<LE)
  	mov		lemask, temp
;
;	Set RAMPZ for CGROM at 0x8000
;
	out		RAMPZ, one			; lpm loads from lower 64kbyte and
								; elpm loads from higher 64kbytes flash
;--------------------------------------------------------------------------
;
;	PORTA	A0..A7
;
;	Multiple Purpose Output Port
;
	out		PORTA, ff
	out		DDRA, ff
;--------------------------------------------------------------------------
;
;	PORTB	INP, KEY, IML, INH, CE, RES, SHLD
;
.equ	VPB	= 0			; Vector Pull from 65xxx
.equ	INP	= 1			; Input Control Signal Enable
.equ	KEY	= 2			; Key Wait / D7 control
.equ	IML = 3			; Initial Machine Load
.equ	INH	= 4			; Inhibit '166 clock
.equ	CE	= 5			; Video RAM Chip Enable
.equ	RES	= 6			; 65xxx reset
.equ	SHLD= 7			; SHLD of 74HC166 connected to OC3B
;
#define	IMLPIN		PORTB, IML
#define RESETPIN	DDRB, RES
;
	ldi		temp, (1<<INP) | (1<<KEY)  | (1<<IML) | (1<<CE)
	out		PORTB, temp
	ldi		temp, (1<<INP) | (1<<KEY) | (1<<IML) | (1<<INH) | (1<<CE) | (1<<SHLD)
	out		DDRB, temp	
;--------------------------------------------------------------------------
;
;	PORTC	D0..D7
;
;	Data Port (vidoe RAM and Input control signals 'PORTF'). This port
;	is normally an input port, only when writing to DPRAM the direction
;	register is set to output but all write routines must reset the
;	direction register to all inputs after they finish writing.
;
	out		DDRC, zero
;--------------------------------------------------------------------------
;
;	PORTD	RXD0, TXD0, DATA, CLR,   CLOCK, HSYNC, SHLD, LE
;
.equ	RX		= 0
.equ	TX		= 1
.equ	DATA	= 2		; PS/2 data line connected to RXD1
.equ	BLANK	= 3		; CLR of 74HC166 active low, blank display
.equ	CLOCK	= 4		; PS/2 clock line connected to XCK1
.equ	HSYNC	= 5		; VGA HSync connected to OC1A
.equ	PHI2	= 6		; 65xxx PHI2 clock using OC2B
.equ	LE		= 7		; PORTE Enable

	out		PORTD, zero
	ldi		temp, (1<<BLANK) | (1<<HSYNC) | (1<<PHI2) | (1<<LE)
	out		DDRD, temp
;--------------------------------------------------------------------------
;
;	PortE:	
;
;	Output PIN Extension 74HCT573 connected to PortA using LE
;	to latch new value into 74HCT573.
;
;	PE0				Bits 0 to 3 are never changed in PORTE register
;	PE1				they are only used to set the high address bits
;	PE2				VA7 to VA10 (note VA11 is connected to PA7 and
;	PE3				is handled together with the lower address bits)
;	PE4				Video RAM R/W Select
;	PE5				Video RAM Output Enable
;	PE6				74AC157 Selector Input Text/GR
;	PE7				VGA Vsync
;
.set	WE		= 4
.set	OE		= 5
.set	SEL		= 6
.set	VSYNC	= 7		; VGA VSync
;
	ldi		temp, (1<<WE) | (1<<OE) | (1<<VSYNC) | (1<<SEL)
	out		PORTA, temp
	sbi		PORTD, LE
	cbi		PORTD, LE		; Latch to output Buffer
	in		PORTE, PORTA	; Keep a copy
;--------------------------------------------------------------------------
;
;	PortF:
;
;	Video Control Signals read via a tri-state buffer connected to PortC.
;	Tri-state buffers are controlled via INP signal. 
;
;	Work in Progress!!!
;
;	PINF0	TEXT (Apple II)
;	PINF1	MIXED (Apple II)
;	PINF2	PAGE2 (Apple II)
;	PINF3	HIRES (Apple II)
;	PINF4	80COL (Apple IIe)
;	PINF5	ALTCHAR (Apple IIe)
;	PINF6	GM0
;	PINF7	GM1
;
;	GM1  GM0  |
;	----------+----------------------------------------------------------
;             |
;	 0    0   |
;	 0    1   |
;	 1    0   |
;	 1    1   |
;             |
;
.set	TEXT	= 0			; 
.set	MIXED	= 1			;
.set	PAGE2	= 2
.set	HIRES	= 3
.set	COL80	= 4
.set	ALTCHAR	= 5
.set	GM0		= 6
.set	GM1		= 7
;
	ldi		temp, (1<<TEXT)	; Startup mode is 40-column Text Display
	mov		PORTF, temp
;--------------------------------------------------------------------------
;
;	USART0
;	
;	Interactive user interface to control the SBC, modify the video
;	settings and downloading images to the SBC. 
;
;	115200baud, 8 data, no parity, 1 stop bit
;
	ldi		temp, HIGH(UBRR_VAL)
	sts		UBRR0H, temp
	ldi		temp, LOW(UBRR_VAL)
	sts		UBRR0L, temp
;
;	Enable receiver and transmitter
;
	ldi		temp, (1<<RXEN0)|(1<<TXEN0)
	sts		UCSR0B, temp
;
;	Set frame format: 8data, 1stop bit, no parity
;
	ldi		temp, (3<<UCSZ00)
	sts		UCSR0C, temp
;--------------------------------------------------------------------------
;
;	USART1
;
;	Set for synchronouos operations and receive only to be used as PS/2 
;	interface. The PS/2 frame format uses 1 start, 8 data, odd parity
;	and one stop bit, the only difference to RS-232 is that the clock
;	is provided by the PS/2 keyboard. This is supported by the AVR as
;	synchronouos mode with external clock using XCK1.
;
	ldi		temp, (0<<UMSEL11) | (1<<UMSEL10) | (1<<UPM11) | (1<<UPM10) | (0<<USBS1) | (1<<UCSZ11) | (1<<UCSZ10) | (0<<UCPOL1)
	sts		UCSR1C, temp
	ldi		temp, (1<<RXEN1) | (0<<TXEN1) | (0<<UCSZ12)
	sts		UCSR1B, temp
;--------------------------------------------------------------------------
;
;	Init variables and statemachines
;
	out		GPIOR0, zero
	sts		keyavr, zero
	rcall	videoinit			; 
	rcall	videostart
	rcall	splashscreen
	clr		Flagnormal
	clr		Flagtoggle
	call	ps2init

	sei						; 
	rjmp	mon
;==========================================================================
;
;	Small Monitor, inspired by the Apple II Monitor
;
;==========================================================================
	.dseg
prompt:		.byte	1
mode:		.byte	1
a1l:		.byte	1
a1h:		.byte	1
a2l:		.byte	1
a2h:		.byte	1
a3l:		.byte	1
a3h:		.byte	1
a4l:		.byte	1
a4h:		.byte	1
a5l:		.byte	1
a5h:		.byte	1
bootstrap:	.byte	1
clibuf:		.byte	80
	.cseg
mon:
							; This is the <entry> of the monitor
							; it expects a serout and a serin routine
							; and an initialised UART.
							; char must be defined as the register that
							; is used to send or receive a character via
							; the UART
	ldi		char, 0x07
	rcall	charout
monz:
	ldi		temp, '*'		; Prompt
	sts		prompt, temp
	rcall	getlnz
	sts		mode,zero
	ldi		yl,low(clibuf)
	ldi		yh,high(clibuf)
nxtitm:
	rcall	getnum			; Try and get a number
	ldi		temp,numchr		; getnum returns with first non digit in 'char'
	ldi		zl,low(2*chrtbl)
	ldi		zh,high(2*chrtbl)
chrsrch:
	dec		temp
	brmi	mon				; All entries done reset to monitor
	lpm		r0,Z+			; First byte of table is not used
	lpm		r0,Z+			; Get second byte
	cp		r0,char			; is it equal to character
	brne	chrsrch			; no try next
	rcall	tosub			; call appropriate action routine
	rjmp	nxtitm			; try to find next command
;
;	Try get a hex number
;
getnum:
	clt						; Assume no digit
	sts		a2l,zero		; Prepare number with 0
	sts		a2h,zero

nxtchr:
	ld		char,Y+			; Get Next Character in Buffer
	ldi		temp,0x30		; 
	eor		temp,char		; Convert 
	cpi		temp,0x0A		; was the character between '0' and '9'
	brlo	dig				; yes so we have a digit
	subi	temp,-0x89		; Convert
	cpi		temp,0xFA
	brsh	dig				; If higher than we have a digit
	ret						; no digit just return with Character in char

dig:
	set						; We have a digit
	andi	temp,0x0F		; Isolate Value in Temp
	mov		r0,temp			; Save value in r0
	lds		temp, a2l		; get current a2l
	swap	temp			; exchange lower and higher bits
	push	temp
	andi	temp, 0xF0		; isolate previous lower bits now higher bits
	or		temp, r0		; copy in lower bits of digit
	sts		a2l, temp		; save new value
	pop		temp			; get saved value
	andi	temp, 0x0F		; get the previous higher bits
	mov		r0, temp		; they form the new lower bits of upper byte
	lds		temp, a2h		; get old upper byte
	swap	temp
	andi	temp, 0xF0		; isolate previous lower bits in higher bits
	or		temp, r0		; copy in previous higher bits of lower byte
	sts		a2h, temp		; save new value
	lds		r0, mode		; get mode
	tst		r0				; no mode
	brne	nxtchr
	lds		r0, a2l			; so copy the value to a1 and a3
	sts		a1l, r0
	sts		a3l, r0
	lds		r0, a2h
	sts		a1h, r0
	sts		a3h, r0
	rjmp	nxtchr
;
;	Execute Action routine for command
;
tosub:
	adiw	Z, (2*numchr)-2	; Offset between table (taken into account
	lpm		xl, Z+			; that we skip the entry with the character)
	lpm		xh, Z+
	movw	Z,X				; Get Address to Z
	lds		r0, mode		; Get current mode
	sts		mode, zero		; Clear next mode
	ijmp					; Execute action routine
;
;	get line terminated with <CR>
;
getlnz:
	ldi		char, 0x0d
	rcall	charout
	ldi		char, 0x0a
	rcall	charout
getln:
	lds		char, prompt		; 
	rcall	charout			; Send Prompt
	ldi		yl, low(clibuf)	; Initialize cli buffer address
	ldi		yh, high(clibuf)
	lds		char, bootstrap
	tst		char
	breq	bckspc
	sts		bootstrap, zero
	ret
bckspc:
	cpi		yl, low(clibuf-1)
	brne	nxtchar
	cpi		yh, high(clibuf-1)
	breq	getlnz
nxtchar:
	rcall	charin			; Get a character
	cpi		char, 0x0A		; ignore line-feeds
	breq	nxtchar			;
	cpi		char, 0x03		; Cancel
	brne	getlnz01
	ldi		char, '^'		; then send a carret
	rcall	charout
	ldi		char, 'C'		; and a capital C
	rcall	charout
	rjmp	getlnz			; restart with getline

getlnz01:
	cpi		char, 0x08
	breq	rubout
	cpi		char, 0x7F
	breq	rubout

	push	char
	cpi		char, 0x0D
	breq	getlnz02
	cpi		char, 0x20
	brsh	getlnz02
	ori		char, 0x40
getlnz02:
	rcall	charout			; we have a valid character so make an echo
	pop		char

	cpi		char, 0x60		; lower case ?
	brlo	addinp
	andi	char, 0x5F		; Perhaps yes make sure we have upper case only
addinp:
	st		Y+, char		; Save Character in line buffer
	cpi		char, 0x0D		; <CR>
	brne	nxtchar			; No get another character
	ldi		char, 0x0A
	rcall	charout
	ret						; Else execute command

rubout:
	ldi		char, 0x08
	rcall	charout			; send rubout character
	ldi		char, ' '		; 
	rcall	charout
	ldi		char, 0x08
	rcall	charout
	sbiw	Y, 1
	rjmp	bckspc
;
;	Core Monitor Routine
;
moncr:
	rcall	monbl1
	pop		r0
	pop		r0
	rjmp	monz

monlt:
	lds		r0, a2l
	sts		a4l, r0
	sts		a5l, r0
	lds		r0, a2h
	sts		a4h, r0
	sts		a5h, r0
	ret

monmove:
	lds		xl, a1l
	lds		xh, a1h
	rcall	load
	lds		xl, a4l
	lds		xh, a4h
	rcall	store
	rcall	nxta4
	brlo	monmove
	ret

monverify:
	lds		xl, a1l
	lds		xh, a1h
	rcall	load
	push	r0

	lds		xl, a4l
	lds		xh, a4h
	rcall	load

	pop		temp
	cp		r0, temp
	breq	monverifyok
	push	r0				; Save value from a4
	push	temp			; Save value from a1
	rcall	pra1
	pop		r0				; Print value from a1
	rcall	prbyte
	ldi		temp, ' '
	mov		char, temp
	rcall	charout
	ldi		temp, '('
	mov		char, temp
	rcall	charout
	pop		r0				; Print value from a4
	rcall	prbyte
	ldi		temp, ')'
	mov		char, temp
	rcall	charout
	
monverifyok:
	rcall	nxta4
	brlo	monverify
	ret	

monbl1:						; 
	cpi		yl, low(clibuf+1)
	brne	monblank
	cpi		yh, high(clibuf+1)	
	breq	monxam8			; Just a <CR> in the clibuffer
monblank:
;;;	brtc	monsetmode		; If last character was no digit just set mode
;;;
	brts	monblank2
	sts		mode, r0		; Restore Mode if no digit
	ret						; 

monblank2:
	mov		temp, r0
	cpi		temp, ':'		; Was Mode = ':'
	brne	monxampm
monstore:					; If Mode was ':' we store characters
	sts		mode, r0		; Keep in Store mode for additional data
	lds		r0, a2l			; Get low byte
	lds		xl, a3l			; Get Pointer
	lds		xh, a3h
	rcall	store
	adiw	X, 1
	sts		a3l, xl			; Save Incremented Pointer for next store
	sts		a3h, xh
	ret

monsetmode:
	sts		mode, char		; Just put the character as mode
	ret

monxam8:					; Examine next Memory up to address = 0 (mod 8)
	lds		temp, a1l		; Set A2 = A1 | 0x0007
	ori		temp, 0x07
	sts		a2l, temp
	lds		temp, a1h
	sts		a2h, temp
monmodchk:
	lds		temp, a1l
	andi	temp, 0x07
	brne	mondataout

monxam:
	rcall	pra1
mondataout:
	ldi		temp, ' '
	mov		char, temp
	rcall	charout
	lds		xl, a1l
	lds		xh, a1h
	rcall	load
	rcall	prbyte
	rcall	nxta1
	brlo	monmodchk
	ret

monxampm:
	sbrs	r0, 0			; + and - are odd, : and . are even ASCII values
	rjmp	monxam			; if : or . goto exam
	lds		temp, a2l
	sbrs	r0, 1			; + = 0x2B and - = 0x2D so - has bit 1 clear
	neg		temp			; take negative value
	lds		r0, a1l
	add		r0, temp
	ldi		temp, 0x0d
	mov		char, temp
	rcall	charout
	ldi		temp, 0x0a
	mov		char, temp
	rcall	charout
	ldi		temp, '='
	mov		char, temp
	rcall	charout
prbyte:
	mov		char, r0
	swap	char
	rcall	prhex
	mov		char, r0
prhex:
	ldi		temp, 0x0F
	and		char, temp
prhexz:
	ldi		temp, 10
	cp		char, temp
	ldi		temp, '0'
	brlo	prbyte01
	ldi		temp, 'A'-10
prbyte01:
	add		char, temp
	rjmp	charout

prntx:
	mov		r0, xh
	rcall	prbyte
prntxl:
	mov		r0, xl
	rcall	prbyte
prblnk:
	ldi		temp, ' '
	mov		char, temp
	rcall	charout
	rcall	charout
	rjmp	charout

pra1:
	lds		xl, a1l
	lds		xh, a1h
prx:
	ldi		temp, 0x0d
	mov		char, temp
	rcall	charout
	ldi		temp, 0x0a
	mov		char, temp
	rcall	charout
	rcall	prntx
	ldi		temp, '-'
	mov		char, temp
	rjmp	charout

nxta4:
	lds		temp, a4l
	inc		temp
	sts		a4l, temp
	brne	nxta1
	lds		temp, a4h
	inc		temp
	sts		a4h, temp
nxta1:
	lds		temp, a1l
	lds		r0, a2l
	cp		temp, r0
	lds		temp, a1h
	lds		r0, a2h
	cpc		temp, r0
	lds		temp, a1l
	inc		temp
	sts		a1l, temp
	brne	rts4b
	lds		temp, a1h
	inc		temp
	sts		a1h, temp
rts4b:
	ret	

monquit:
	pop		r0
	pop		r0
	ret
;==========================================================================
;
;	Valid command characters and action routine table
;	Note that up to verify they are part of the core
;	action routines, add your routine after the core
;	routines
;
chrtbl:
	.db		0, ' '		; Blank
	.db		0, '.'		; Range
	.db		0, ':'		; Store
	.db		0, '-'		; Subtract
	.db		0, '+'		; Add
	.db		0, '<'		; Destination
	.db		0, 0x0D		; <CR>
	.db		0, 'M'		; Move
	.db		0, 'V'		; Verify
	.db		0, '?'		; Help Screen
;
;	Video Control uses all ctrl character. We cannot use 0x0A (linefeed),
;	0x03 (cancel), 0x08 (backspace), 0x0D (carriage return).
;	Once video settings are known this will no longer be used.
;
	.db		0, 0x14		; ^T Create Test Pattern
	.db		0, 0x0F		; ^O shldoff
	.db		0, 0x10		; ^P inhpolarity
	.db		0, 0x0C		; ^L hline
	.db		0, 0x12		; ^R dstart
	.db		0, 0X1A		; ^Z Set 40/80
	.db		0, 0x01		; ^A Altchar (toggle)
	.db		0, 0x09		; ^I Info
	.db		0, 0x0E		; ^N Parallel keyboard
	.db		0, 0x16		; ^V Video on/off
;
;	Boot control
;
	.db		0, 0x05		; Erase
	.db		0, 'R'		; Reset
	.db		0, 'W'		; Cycle
	.db		0, 'S'		; VPB status
	.db		0, 'P'		; Toggle PHI2
	.db		0, 'K'		; Clock
	.db		0, 'T'		; Stop
	.db		0, 'X'		; XModem
	.db		0, 'Y'		; Load default ROM
	.db		0, 'I'		; Toggle IML
	.db		0, 'O'		; Output

	.equ	chrtbll = PC-chrtbl
;
;
;
subtbl:
	.equ	numchr	= subtbl-chrtbl
	.dw		monblank
	.dw		monsetmode
	.dw		monsetmode
	.dw		monsetmode
	.dw		monsetmode
	.dw		monlt
	.dw		moncr
	.dw		monmove
	.dw		monverify
	.dw		splashscreen
;
	.dw		tstpattern	; ^T
	.dw		setshldoff	; ^O
	.dw		setinhpol	; ^P
	.dw		sethline	; ^L 
	.dw		setdstart	; ^R 
	.dw		set4080		; ^Z 
	.dw		taltchar	; ^A
	.dw		printinfo	; ^I 
	.dw		prtps2		; ^N
	.dw		setvideo	; ^V
;
	.dw		imlerase	; ^E
	.dw		imlreset	; 'R'
	.dw		imlcycle	; 'W'
	.dw		imlstatus	; 'S'
	.dw		imltoggle	; 'P'
	.dw		imlclock	; 'K'
	.dw		imlstop		; 'T'
	.dw		imlxmodem	; 'X'
	.dw		imldefault	; 'Y'
	.dw		imliml		; 'I'
	.dw		imloutput	; 'O'

	.equ	subtbll = PC-subtbl

	.if subtbll != chrtbll
	.error "Char- and Sub- table are not equally sized"
	.endif

charout:
	jmp		serout

charin:
	jmp		serin

	.include	"iml.asm"

load:
	cpi		xh, 0x40
	brsh	loaddpram
	ld		r0, X
	ret

store:
	cpi		xh, 0x40
	brsh	storedpram
	st 		X, r0
	ret
;
;	Load and Store to dualport RAM, the mapping from the
;	video display routine is A0..A6, A11, A7..A10.
;	A11 maps to MAIN/AUX of the Apple II
;
.macro	dpmap
	lsl		xl				; Put A0..A6 to bits 1..7
	rol		xh				; Get A7..A11 to bits 0..4
	lsr		xl
	bst		xh, 4
	bld		xl, 7	
.endmacro

.macro	vstop
	rcall	videostop
.endmacro
.macro	vstart
	push	r0
	rcall	videostart
	pop		r0
.endmacro

loaddpram:
	push	xl
	push	xh
	dpmap

	vstop

	andi	xh, 0x0F
	or		xh, PORTE
	out		PORTA, xh

	cbi		PORTA, OE		; output enable
	sbi		PORTD, LE		; 
	nop
	nop
	cbi		PORTD, LE		; store to 74HC573
	out		PORTA, xl		; bits A0..A7
	cbi		PORTB, CE		; enable RAM
	nop
	nop
	nop
	in		r0, PINC		; read RAM
	sbi		PORTB, CE		; disable RAM
	out		PORTA, PORTE	; restore PORTE
	sbi		PORTD, LE
	nop
	nop
	cbi		PORTD, LE
	
	vstart
	
	pop		xh
	pop		xl
	ret

storedpram:
	push	xl
	push	xh
	dpmap

	vstop

	out		DDRC, ff				; outputs
	andi	xh, 0x0F
	or		xh, PORTE
	out		PORTA, xh
	cbi		PORTA, WE
	sbi		PORTD, LE
	nop
	nop
	cbi		PORTD, LE
	out		PORTA, xl
	cbi		PORTB, CE
	out		PORTC, r0
	nop
	nop
	sbi		PORTB, CE
	out		DDRC, zero				; inputs
	out		PORTA, PORTE
	sbi		PORTD, LE
	nop
	nop
	cbi		PORTD, LE

	vstart

	pop		xh
	pop		xl
	ret
;
;	Toggle ATLCHAR flag
;
taltchar:
	ldi		temp, (1<<ALTCHAR)
	eor		PORTF, temp
	ret
;
;	create a test pattern 0x00...0xFF at 0x1000 in AVR RAM
;
tstpattern:
	ldi		xh, 0x10
	clr		xl
tst010:
	mov		r0, xl
	st		X+, r0
	tst		xl
	brne	tst010
	ret
;
;	set 40 column mode
;
set4080:
	rcall	videostop
	lds		temp, a1l
	sbrc	temp, 0
	rjmp	set80

	ldi		temp, (1<<COM0B1)
	sts		inhcon, temp
	ldi		zl, low(hline_40c)
	ldi		zh, high(hline_40c)
	sts		nexthstate, zl
	sts		nexthstate+1, zh
	rjmp	videostart
;
;	set 80 column mode
;
set80:
	sts		inhcon, zero
	ldi		zl, low(hline_80c)
	ldi		zh, high(hline_80c)
	sts		nexthstate, zl
	sts		nexthstate+1, zh
	rjmp	videostart
;
;	set hline
;
sethline:
	rcall	videostop
	lds		temp, a1l
	sts		hline, temp
	lds		temp, a1h
	sts		hline+1, temp
	rjmp	videostart
;
;	set start of hline display interrupt
;
setdstart:
	rcall	videostop
	lds		temp, a1l
	sts		dstart, temp
	lds		temp, a1h
	sts		dstart+1, temp
	rjmp	videostart
;
;	set offset of SHLD to hline timer
;
setshldoff:
	rcall	videostop
	lds		temp, a1l
	sts		shldoff, temp
	rjmp	videostart
;
;	change inh polarity
;
setinhpol:
	rcall	videostop
	lds		temp, a1l
	andi	temp, 0x01
	sts		inhpol, temp
	rjmp	videostart
;
;	Manually stop/start video
;
setvideo:
	lds		temp, a1l
	sbrs	temp, 0
	rjmp	videostop
	rjmp	videostart
;
;	enable/disable printing of scancodes
;
prtps2:
	lds		temp, a1l
	andi	temp, 0x01
	sts		keyavr, temp
	ret
;
;	print current timer values
;
	.org	PC+99
printinfo:
	ldi		char, 'h'
	rcall	serout
	ldi		char, 'l'
	rcall	serout
	ldi		char, 'i'
	rcall	serout
	ldi		char, 'n'
	rcall	serout
	ldi		char, 'e'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, hline+1
	rcall	prbyte
	lds		r0, hline
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 'h'
	rcall	serout
	ldi		char, 's'
	rcall	serout
	ldi		char, 'y'
	rcall	serout
	ldi		char, 'n'
	rcall	serout
	ldi		char, 'c'
	rcall	serout
	ldi		char, 'w'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, hsyncw+1
	rcall	prbyte
	lds		r0, hsyncw
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 'd'
	rcall	serout
	ldi		char, 's'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, 'a'
	rcall	serout
	ldi		char, 'r'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, dstart+1
	rcall	prbyte
	lds		r0, dstart
	rcall	prbyte
	rcall	seroutcrlf
	
	ldi		char, 's'
	rcall	serout
	ldi		char, 'h'
	rcall	serout
	ldi		char, 'l'
	rcall	serout
	ldi		char, 'd'
	rcall	serout
	ldi		char, 'o'
	rcall	serout
	ldi		char, 'f'
	rcall	serout
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, shldoff
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 'i'
	rcall	serout
	ldi		char, 'n'
	rcall	serout
	ldi		char, 'h'
	rcall	serout
	ldi		char, 'p'
	rcall	serout
	ldi		char, 'o'
	rcall	serout
	ldi		char, 'l'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, inhpol
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 'p'
	rcall	serout
	ldi		char, 'x'
	rcall	serout
	ldi		char, 'l'
	rcall	serout
	ldi		char, 'p'
	rcall	serout
	ldi		char, 'c'
	rcall	serout
	ldi		char, 'h'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, pxlpch
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 't'
	rcall	serout
	ldi		char, '0'
	rcall	serout
	ldi		char, 'c'
	rcall	serout
	ldi		char, 'n'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, t0info
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 't'
	rcall	serout
	ldi		char, '1'
	rcall	serout
	ldi		char, 'c'
	rcall	serout
	ldi		char, 'n'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, t1info+1
	rcall	prbyte
	lds		r0, t1info
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 't'
	rcall	serout
	ldi		char, '3'
	rcall	serout
	ldi		char, 'c'
	rcall	serout
	ldi		char, 'n'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, t3info+1
	rcall	prbyte
	lds		r0, t3info
	rcall	prbyte
	rcall	seroutcrlf

	ldi		char, 'j'
	rcall	serout
	ldi		char, 'i'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, 't'
	rcall	serout
	ldi		char, 'e'
	rcall	serout
	ldi		char, 'r'
	rcall	serout
	ldi		char, '.'
	rcall	serout
	ldi		char, ':'
	rcall	serout
	lds		r0, jtcnt1+1
	rcall	prbyte
	lds		r0, jtcnt1
	rcall	prbyte
	rcall	seroutcrlf
	cbi		GPIOR0, JITTER

	ret

;==========================================================================

	.include	"xmodem-receive-1-4.3.asm"
	.include	"ps2statemachine.asm"



;--------------------------------------------------------------------------
;
;	OC1B Match Interrupt
;
;	The VGA signal generation is implemented as a state-machine. Every
;	OCR1B interrupt we fetch the current state and jump to the routine
;	that processes the state.
;
;	hline_80c	generate a scan line in 80-column mode
;	hline_40c	generate a scan line in 80-column mode
;	vfporch_d	generate vertical front porch lines, actually only
;				performs a count down and when it reaches the end
;				sets VSYNC to low
;	vsync_d		counts down the VSYNC and at the end sets VSYNC to high
;	vbporch_d	generate vertical back porch lines, actually only
;				performs a count down and when it reaches the end
;				sets the appropriate scan line routine (hline_80c or
;				hline_40c) as next state.
;
;	Future states will include support for low-res graphics and mixed
;	display.
;
;	Caution: We only have 700 cycles per ISR fo which 560 are used by the
;	display generation, so every cycle counts!
;
ISR_hline:						;  8	push pc and execute jmp at vector
	isave						; 11	save common registers and SREG
	lds		zl, nextstate		;  2
	lds		zh, nextstate+1		;  2
	ijmp						;  2	-> 25 cycles
;
;	Dejitter start of display code due to variable cycle length, we use 
;	TCNT1 which triggered an interrupt when 'dstart' was reached. 
;	depending on the number of cycles the instruction we interrrupted
;	takes TCNT1 has reached a different value. When we assume that the
;	interrupt is served immediately then TCNT1 counted 25 cycles the
;	moment we read the value. If the interrupt required the current
;	instruction to execute, then TCNT1 is incremented the number of
;	cycles this instruction required. So we take TCNT1 minus 'dstart'
;	and minus 25. The result will be the number of additional cycles
;	were required before the interrupt is served. We use this to skip
;	this count of nop's. The result is that the instructions of the
;	interrupt routine are always executed with the same value TCNT1
;	respectively. That also means the output of vidoe data is always
;	at the same position relative to the HSYNC pulse which also 
;	depends on TCNT1. Therefore we should not see any jitter in the VGA
;	screen.
;
;	NOTE: this de-jitter logic requires that interrupts are never 
;	disabled, as this might delay the entry even more and that is
;	not taken into account here. So when you need to do something
;	in the main program, that does not allow interrupts, just disable
;	the match OCR1 interrupt as this is the only interrupt active.
;
	.equ	jt	= 25
	.macro	hlineadjust
	lds		zl, TCNT1L			;  2	dstart+25+...
	lds		zh, TCNT1H			;  2
	lds		xl, dstart			;  2
	lds		xh, dstart+1		;  2
	sub		zl, xl				;  1
	sbc		zh, xh				;  1	25+...				
	subi	zl, low(jt-hla)		;  1	-25+hla:
	sbci	zh, high(jt-hla)	;  1
	ijmp						;  2
hla:
	nop							;  1
	nop							;  1	1 cycle, most instructions
	nop							;  1	2 cycle, mostly branches
	nop							;  1	3 cycles, jumps
	nop							;  1	4 cycles, jumps and calls
	nop							;  1	5 cycles, RET and RETI
	nop							;  1	+ safety cycles
	nop							;  1	-> 22 cycles 
	.endmacro
;
;	Check keyboard, the KWAIT flag is located in GPIOR0 and the KBUF
;	is located in GPIOR1
;
	.macro	checkkbd
	sbis	GPIOR0, KWAIT		; 1/1
	rjmp	nokey				; 1/2
	in		xl, GPIOR1			; 1
	out		PORTA, xl			; 1
	cbi		PORTB, KEY			; 2
	cbi		GPIOR0, KWAIT		; 2
	sbi		PORTB, KEY			; 2
nokey:							;		-> 10/3 cycles
	.endmacro
;
;	In this version of VGA controller PORTA is used for both, the
;	low byte address of the video RAM and the bitmap for the shift
;	register. We time-multiplex these functions. During 6 cycles we
;	use it as the low-byte address and for 1 cycle we use it to send
;	the bitmap ot the shift register.
;
;	The following block must be in sync with SHLD. First the BLANK
;	signal must be deasserted (which is connected to !CLR of the
;	74HCT166) and activate the shift register, followed by outputting
;	a zero bit pattern so that the first thing the shift register loads
;	is all zero bits and shifts zero bits out and not was has been
;	previously in PORTA, as at this stage we do not have a bitmap of
;	character code from the video RAM.
;
;	The DP RAM has a certain access time, but we have no time to
;	wait for the DP RAM. Therefore we interleave setting the video
;	RAM address and translation of video data to bitmap. First we
;	set the video address of the first byte, and wait for the DP RAM
;	and then retrieve the character code. In 80 Column mode this
;	is in AUX RAM (A11, aka R16 set) but just output 'zero' as the
;	bitmap which is equivalent to blanking the display. Then we continue
;	with putting the next address and during the time we need to wait
;	for the DP RAM we calculate the next address and translate the 
;	previously retrieved character code to a bitmap. Every next address
;	toggles A11/R16 and on every other step it incrementes A0..A6. The
;	hardware is mapped in a way that all these addresses are connected
;	to PORTA, so managing the addresses of a display line only involves
;	updating a single byte. We use 'subi' that comes very handy for
;	calculating the next address. As there is no overflow from A0..A6
;	when incrementing the display line address bits, by selecting the
;	constant we subtract we can individually toggle A11 or toggle A11
;	and increment A0..A6 at the same time.
;
;
;--------------------------------------------------------------------------
;
;	Vertical Line	80
;
;	Each display state must start with a anti-jitter landing code. This
;	code is nothing else than a number of 'nop' instructions. The number
;	must match the maximum offset we might add in ISR_hline.
;
;	We need to dejitter the variable number of cycles  a interrupted
;	intstruction may have. Instructions take from 1 to 5 cycles. 
;
;	Note the landing zone only accounts for variable cycles of a single
;	instructions. This means that we no longer may disable interrupts in
;	the main code. Instead the main code must disable the hline interrupt
;	and resume it afterwards.
;
hline_80c:
;
;	De-jitter must be the first action
;
	hlineadjust						;		-> 22
;
;	Prepare CGROM pointer. scan runs from 0 to 19 as we have 20 scan
;	lines per character line. The CGROM normally have 8 lines. Hence
;	we first ranslate the scan line to a offset in CGROM. The CGROM
;	is organised as 8 blocks of 256bytes, each block corresponds the
;	the nth bitmap line of the 256 byte values.
;
	scanto	zh						;  7
	subi	zh, high(-(2*cgromp80))	;  1
;
;	The CGROM has 3 sections. The first two sections contain
;	the default character set in two version. The difference between
;	these two versions is that 64 characters out of the 256 characters
;	are inverted to support flashing characters in hardware for 
;	those 64 character codes. 'flash' is incremented when VSYNC is 
;	generated, that is every 60ms. If bit 5 is set we will set the
;	second section.
;
	sbrc	flash, 5				;  1
	ori		zh, high(2*(cgromi80-cgromp80))
									;  1	Offset to inverse bitmap
;
;	The third section has an secondary character set in CGROM which does
;	not support hardware flash. Therefore if the secondary character set 
;	is selected we advance the pointer to the alternate character set and
;	revert an eventually added offset to the second section.
;
	sbrc	PORTF, ALTCHAR			;  1
	ori		zh, high(2*(cgroma80-cgromp80))
									;  1	Offset to secondaray charset
	sbrc	PORTF, ALTCHAR			;  1
	andi	zh, ~high(2*(cgromi80-cgromp80))
									;  1	Revert offset to inverse bitmap
;
;	The video RAM is organzied as in a Apple IIe. The advantage of this
;	mapping is that all addresses of a character display line have the
;	same high byte address (i.e. A8..A15 are fixed). The Apple IIe packs
;	three lines into one chunk of 128bytes, i.e. 3x40characters. However
;	the lines on the screen and in memory are interleaved. The first 128
;	bytes are used for the lines 1,9,17, then ext are used for lines 2,
;	10, 18 etc. This is perfect as we only have 7 cycles during display
;	per character, enough to set the lower 8 address bits, convert the
;	character code to bitmap and output the bitmap for the shift register
;	to read, but not enough time to set higher 8 address bits.
;
;	The 'linetox' macro takes the 'line' counter and creates the Apple II
;	compatible base address (see BASCALC in Apple Monitor) and buts the
;	address into xl:xh. Note that A11 of the DP-RAM is connected to PA7.
;	This is to support the 80-column Apple IIe dislay. We assume that 
;	A11 of the other side of the DP-RAM is connected to the MAIN/AUX 
;	RAM select signal (aka R16)
;
	linetox							;  8	-> 22
;
;	Now we have a 16-bit video RAM base address for the character line we
;	want to display. First we combine the address bits A7..A10 with the
;	rest of the PORTE bits (remember PORTE is a 74HC573 we use as a port
;	expander), especially we assert /OE to read from the DP-RAM
;
	or		xh, PORTE			;  1	Combine bits and addresses set the
	out		PORTA, xh			;  1	port that connects to the inputs
	cbi		PORTA, OE			;  2	of the port expander, assert /OE.
	out		PIND, lemask		;  1	Then latch the value into the
	cbi		PORTB, CE			;  2	74HC573. And interleave it with
	out		PIND, lemask		;  1	asserting /CE for a longer pulse.



	subi	xl, 0x80			;  1	AUX Memory for first column
	out		PORTA, xl			;  1	Set Low Byte Address
	subi	xl, 0x80			;  1	Toggle Bit 7
	sbi		PORTD, BLANK		;  2	Enable Display
	in		zl, PINC			;  1	Get value of first byte
;
;	The following instruction must be in sync with SHLD = Low!!!
;
	out		PORTA, zero			;  1	Set bitmap to zero for first shift
								;		-> 15
;	
;	Macro to display two characters in 80 column mode, must be exactly
;	14 cycles and the video data must be output at cycle 7 and 14!
;
.macro		char80				;		Put AUX and MAIN Character
	out		PORTA, xl			;  1*	Set Low Byte Address of next byte
	subi	xl, 0x7F			;  1*	Toggle Bit 7 and increment Bit0:6
	elpm	vd, Z				;  3*	Translate previous byte to bitmap
	in		zl, PINC			;  1*	Get next byte
	out		PORTA, vd			;  1*	Set bitmap of previous for shift
	out		PORTA, xl			;  1*	Set Low Byte Address of next byte
	subi	xl, 0x80			;  1*	Toggle Bit 7
	elpm	vd, Z				;  3*	Translate previous byte to bitmap
	in		zl, PINC			;  1*	Get next byte
	out		PORTA, vd			;  1*	Set bitmap of previous for shift
.endmacro
	char80						; 01	Set address of next
	char80						; 02
	char80						; 03
	char80						; 04
	char80						; 05
	char80						; 06
	char80						; 07
	char80						; 08
	char80						; 09
	char80						; 10

	char80						; 11
	char80						; 12
	char80						; 13
	char80						; 14
	char80						; 15
	char80						; 16
	char80						; 17
	char80						; 18
	char80						; 19
	char80						; 20

	char80						; 21
	char80						; 22
	char80						; 23
	char80						; 24
	char80						; 25
	char80						; 26
	char80						; 27
	char80						; 28
	char80						; 29
	char80						; 30

	char80						; 31
	char80						; 32
	char80						; 33
	char80						; 34
	char80						; 35
	char80						; 36
	char80						; 37
	char80						; 38
	char80						; 39
	char80						; 40

	out		PORTA, PORTE		;  1
	out		PIND, lemask		;  1
	cbi		PORTD, BLANK		;  2
	out		PIND, lemask		;  1

	sbi		PORTB, CE			;  2
	
	inc		scan				;  1
	cpi		scan, 20			;  1
	brlo	nextscan80c			;  1
	inc		line				;  1
	cpi		line, 24			;  1
	brlo	nextline80c			;  1

;--------------------------------------------------------------------------
;
;	End of display, set next state is vertical front porch
;
	ldi		line, vfplines		;  1	Vertical Front Porch length

	ldi		zl, low(vfporch_d)	;  1
	ldi		zh, high(vfporch_d)	;  1
	sts		nextstate, zl		;  2
	sts		nextstate+1, zh		;  2

	irestore					; 11
	reti						;  5	
;--------------------------------------------------------------------------
;
;	Display next scan line
;

	irestore
	reti

;--------------------------------------------------------------------------
;
;	Display next line
;
nextline80c:
	clr		scan
;--------------------------------------------------------------------------
;
;	Display next scan line
;
nextscan80c:
	checkkbd
	irestore
	reti


;--------------------------------------------------------------------------
;
;	Vertical Line	40
;
;	Each state must start with a anti-jitter landing code. This code is
;	nothing else than a number of 'nop' instructions. The number must
;	match the maximum offset we might add in ISR_hline.
;
hline_40c:
;
;	For comments, see hline_80c
;
	hlineadjust						; 		-> 22
;
	scanto	zh						;  7
	subi	zh, high(-(2*cgromp80))	;  1
	sbrc	flash, 5				;  1
	ori		zh, high(2*(cgromi80-cgromp80))
									;  1	Offset from primary to inverse
	sbrc	PORTF, ALTCHAR			;  1
	ori		zh, high(2*(cgroma80-cgromp80))
									;  1	Offset from primary to alternate
	sbrc	PORTF, ALTCHAR			;  1
	andi	zh, ~high(2*(cgromi80-cgromp80))
									;  1	Alternate has no flashing set
	linetox							;  8	-> 22
;
;
;
	or		xh, PORTE			;  1	Combine bits and addresses set the
	out		PORTA, xh			;  1	port that connects to the inputs
	cbi		PORTA, OE			;  2	of the port expander, assert /OE.
	out		PIND, lemask		;  1	Then latch the value into the
	cbi		PORTB, CE			;  2	74HC573. And interleave it with
	out		PIND, lemask		;  1	asserting /CE for a longer pulse.
	ori		xl, 0x80			;  1	Select MAIN RAM
;
;	The following block must be in sync with SHLD, that is the BLANK
;	signal must be deasserted (which is connected to !CLR of the
;	74HCT166) and activate the shift register, followed by outputting
;	a zero bit pattern so that the first thing the shift register loads
;	is all zero bits and shifts zero bits out and not was has been
;	previously in PORTA
;

	sbi		PORTD, BLANK		;  2
	out		PORTA, zero			;  1	-> 12
;
;	Macro that puts one character in 40 column mode. The macro must
;	take exactly 14 cycles adn the last cycle must place the video
;	data on PORTA
;
;
;
.macro		char40
	out		PORTA, xl			;  1*
	inc		xl					;  1*
	nop	 						;  1*
	nop	 						;  1*
	nop	 						;  1*
	in		zl, PINC			;  1*
	elpm	vd, Z				;  3*
	out		PORTA, vd			;  1*
	nop	 						;  1*
	nop	 						;  1*
	nop	 						;  1*
	nop	 						;  1*
.endmacro
	char40						; 01
	char40						; 02
	char40						; 03
	char40						; 04
	char40						; 05
	char40						; 06
	char40						; 07
	char40						; 08
	char40						; 09
	char40						; 10

	char40						; 11
	char40						; 12
	char40						; 13
	char40						; 14
	char40						; 15
	char40						; 16
	char40						; 17
	char40						; 18
	char40						; 19
	char40						; 20

	char40						; 21
	char40						; 22
	char40						; 23
	char40						; 24
	char40						; 25
	char40						; 26
	char40						; 27
	char40						; 28
	char40						; 29
	char40						; 30

	char40						; 31
	char40						; 32
	char40						; 33
	char40						; 34
	char40						; 35
	char40						; 36
	char40						; 37
	char40						; 38
	char40						; 39
	char40						; 40

	nop							;  1
	nop							;  1
	nop							;  1
	nop							;  1
	nop							;  1

	sbi		PORTB, CE			;  2
	out		PORTA, PORTE		;  1
	sbi		PORTD, LE			;  2
	cbi		PORTD, LE			;  2

	cbi		PORTD, BLANK		;  2
	
	inc		scan				;  1
	cpi		scan, 20			;  1
	brlo	nextscan40c			;  1
	inc		line				;  1
	cpi		line, 24			;  1
	brlo	nextline40c			;  1

;--------------------------------------------------------------------------
;
;	End of display, set next state is vertical front porch
;
	ldi		line, vfplines	; Vertical Front Porch length

	ldi		zl, low(vfporch_d)	;  1
	ldi		zh, high(vfporch_d)	;  1
	sts		nextstate, zl		;  2
	sts		nextstate+1, zh		;  2

	irestore					; 11
	reti						;  4	-> 41
;--------------------------------------------------------------------------
;
;	Display next line
;
nextline40c:
	clr		scan
;--------------------------------------------------------------------------
;
;	Display next scan line
;
nextscan40c:
	checkkbd
	irestore
	reti


;==========================================================================
;
;	Vertical Front Porch
;
;	Each state must start with a anti-jitter landing code. This code is
;	nothing else than a number of 'nop' instructions. The number must
;	match the maximum offset we might add in ISR_hline.
;
vfporch_d:
	dec		line
	brne	nextvfporch
	ldi		line, vsynclines	; Next State is VSYNC
	ldi		zl, low(vsync_d)
	ldi		zh, high(vsync_d)
	sts		nextstate, zl
	sts		nextstate+1, zh
	out		PORTA, PORTE
	cbi		PORTA, VSYNC
	sbi		PORTD, LE
	cbi		PORTD, LE
	in		PORTE, PORTA
nextvfporch:
	checkkbd
	irestore
	reti

;--------------------------------------------------------------------------
;
;	Vertical Sync
;
;	Each state must start with a anti-jitter landing code. This code is
;	nothing else than a number of 'nop' instructions. The number must
;	match the maximum offset we might add in ISR_hline.
;
vsync_d:

	lds		zl, TCNT1L			;  2	dstart+25+...
	lds		zh, TCNT1H			;  2
	sbis	GPIOR0, JITTER
	sts		jtcnt1, zl
	sbis	GPIOR0, JITTER
	sts		jtcnt1+1, zh
	sbi		GPIOR0, JITTER
	dec		line
	brne	nextvsync
	lds		xl, TCNT3L
	lds		xh, TCNT3H
	lds		zl, TCNT1L
	lds		zh, TCNT1H
	sub		zl, xl
	sbc		zh, xh
	sts		tmrskew, zl
	sts		tmrskew+1, zh
	ldi		zl, low(vbporch_d)
	ldi		zh, high(vbporch_d)
	sts		nextstate, zl
	sts		nextstate+1, zh
	out		PORTA, PORTE
	sbi		PORTA, VSYNC
	sbi		PORTD, LE
	cbi		PORTD, LE
	in		PORTE, PORTA
	inc		flash
nextvsync:
	checkkbd
	irestore
	reti
;--------------------------------------------------------------------------
;
;	Vertical Back Porch, also used as the entry for the state machine
;
;	Each state must start with a anti-jitter landing code. This code is
;	nothing else than a number of 'nop' instructions. The number must
;	match the maximum offset we might add in ISR_hline.
;
vbporch_d:
	linetox
	inc		line
	cpi		line, vbplines		; Next State vertical back Porch
	brne	nextvbporchd

	clr		scan
	clr		line
	lds		zl, nexthstate
	lds		zh, nexthstate+1
	sts		nextstate, zl
	sts		nextstate+1, zh

	checkkbd
	irestore
	reti

nextvbporchd:	
	checkkbd
	irestore
	reti

;==========================================================================


romimage:
	.include "A2ROM.inc"
romimage1:
	.include "APPLE2PLUS.inc"
	
;--------------------------------------------------------------------------
;
;	Character ROM has been placed into the upper half of the FLASH and
;	must be addressed using ELPM instruction with RAMPZ set to 0x01.
;
.include "characterromIV.inc"
