; XMODEM/CRC Receiver for AVR
;
; A simple file transfer program to allow upload from a console device
; to the SBC utilizing the x-modem/CRC transfer protocol.  
;
;**************************************************************************
; This implementation of XMODEM/CRC does NOT conform strictly to the 
; XMODEM protocol standard in that it (1) does not fall back to the Checksum
; mode. To correctly time the characters it uses Timer 3 and assumes that
; F_CPU is set to the frequency and that USART0 is configured and used.
;
; (1) Most modern terminal programs support XMODEM/CRC which can detect a
; wider range of transmission errors so the fallback to the simple checksum
; calculation was not implemented to save space.
;**************************************************************************
;
;-------------------------- The Code ----------------------------
;
; Register Usage
;
;	This code makes use of the three 16-bit pointesr X, Y and Z and two
;	registers that must be named 'char' and 'temp' and must be registers
;	above r15. In addition it assumes that the two registers 'zero' and
;	'one' are defined a set with the values 0x00 and 0x01.
;
;
; XMODEM Control Character Constants
;
.equ	SOH		=	0x01		; start block
.equ	EOT		=	0x04		; end of text marker
.equ	ACK		=	0x06		; good block acknowledged
.equ	NAK		=	0x15		; bad block acknowledged
.equ	CAN		=	0x18		; cancel (not standard, not supported)
.equ	CR		=	0x0d		; carriage return
.equ	LF		=	0x0a		; line feed
.equ	ESC		=	0x1b		; ESC to exit

;^^^^^^^^^^^^^^^^^^^^^^ Start of Program ^^^^^^^^^^^^^^^^^^^^^^
;
; Xmodem/CRC upload routine based on the routine
; by Daryl Rictor, Oct 30, 2012
;
; v 1.2	Peter Schranz, 18-Nov-2013 various changes
;		- 	WritePage subroutine that contains all SPM actions
;		-	Now expects the first two bytes of the file to contain
;			the start address, so you can use the binaries from TASS
;		-	only lower 64kbytes of Flash are supported the engine
;			is above 0x8000 and should be loaded via ISP
;
; v 1.4 Peter Schranz 29-May-2015
;		-	Timed Receive Routine
;		-	Minimize use of registers
;		-	Save X and Y Registers
;		-	Interface to 6502 ROM-Less Bootstrap
;		-	Return on Error and do not jump to "reset"
;		-	Assumese the following registers being defined
;			zero	contains 0x00
;			one		contains 0x01
;			temp	temporary register with imm. support
;			char	temporary register with imm. support
;
; v 1.4.2 Peter Schranz 20-08-2015
;		-	Introduce Options for various envrionments
;		-	Add AVR-IML support
;
; v 1.4.3 Peter Schranz 14-09-2015
;		-	VGA Controller IV IML integration
		.dseg
		align	8
rambuff:.byte	256				; should be 256 aligned
recbuff:.byte	132				; must be 256 aligend
blkno:	.byte	1
bflag:	.byte	1
crc:	.byte	2
zptr:	.byte	2
		.cseg
		
XModemAUX:
		set						; Load to 2nd ROM Bank
		brts	XModemStart
imlxmodem:
XModem:
		clt						; Load to 1st ROM Bank

XModemStart:
		push	r24
		push	r25
		push	xl
		push	xh
		push	yl
		push	yh
		call	videostop

		ldi		xl, low(RamBuff)
		ldi		xh, high(RamBuff)	; Temp RAM page buffer
		ldi		yh, high(RecBuff)
		rcall	PrintMsg		; send prompt and info
		sts		blkno, one		; set block # to 1
		sts		bflag, zero
StartCrc:	
		ldi		char, 'C'		; "C" start with CRC mode
		rcall	Put_Chr			; send it
		ldi		char, 3			; 3 seconds time-out, xmodem standard is 10.
		rcall	GetByte			; wait for input
       	brcs	GotByte			; byte received, process it
		brcc	StartCrc		; resend "C"
;
;
;
StartBlk:	
		ldi		char, 1
		rcall	GetByte			; get first byte of block
		brcc	StartBlk		; timed out, keep waiting...
GotByte:
		cpi		char, ESC		; quitting?
		brne	GotByte1		; no
		jmp		xexit			; YES - exit
GotByte1:
		cpi		char, SOH		; start of block?
		breq	BegBlk			; yes
		cpi		char, EOT		;
		brne	BadCrc			; Not SOH or EOT, so flush buffer & send NAK	
		rjmp	Done			; EOT - all done!
;
;		SOH - Start of Header received
;
;		Get the record and then check if everything is ok
;
BegBlk:		
		clr		yl				; byte counter
GetBlk:		
		ldi		char, 1			; 1 seconds time-out, xmodem standard is 1.
		rcall	GetByte			; get next character
		brcc	BadCrc			; chr rcv error, flush and send NAK
		st		Y+, char		; (buff) good char, save it in the rcv buffer
		cpi		yl, 0x84		; <01> <FE> <128 bytes> <CRCH> <CRCL>
		brne	GetBlk			; get 132 characters
;
;		A record has a total of 132 bytes. The first byte contains the
;		blocknumber, the second byte contains the complement of the
;		blocknumber. Then follow 128 data bytes. The last two bytes
;		contain the checksum, in our case we only support CRC-16.
;
		clr		yl				;
		ld		char, Y+		; (buff) get block # from buffer
		lds		temp, blkno
		cp		char, temp		; compare to expected block #	
		breq	GoodBlk1		; matched!
		rcall	Print_Err		; Unexpected block number - abort	
		rcall	Flush			; mismatched - clean-up
		jmp		xexit			; unexpected block # - exit
;
;		Received expected block number
;
GoodBlk1:	
		com		temp
		ld		char, Y+
		cp		char, temp		; compare with expected 1's comp of block #
		breq	GoodBlk2 		; matched!
		rcall	Print_Err		; Unexpected block number - abort	
		rcall 	Flush			; mismatched - flush buffer and then do BRK
		jmp		xexit			; bad 1's comp of block#	
;
;		and the second byte is the correct complement of the blocknumber.
;		The checksum is only calculated from the 128 data bytes. First we
;		initialise the checksum and then we calculate the checksum and last
;		we compare it to the checksum received
;
GoodBlk2:	
		sts		crc, zero		; init CRC
		sts		crc+1, zero
		ldi		yl, 0x02		; data starts with the 3rd byte
CalcCrc:	
		ld		char, Y+		; calculate the CRC for the 128 bytes of data	
		rcall	UpdCrc			; could inline sub here for speed
		cpi		yl, 0x82		; 128 bytes (i.e. pointer at 130)
		brne	CalcCrc			;
		ld		char, Y+		; get hi CRC from buffer
		lds		temp, crc+1
		cp		char, temp
		brne	BadCrc			; bad crc, send NAK
		ld		char, Y			; get hi CRC from buffer
		lds		temp, crc
		cp		char, temp
		breq	GoodCrc			; good CRC
;
;		CRC does not match, flush the receive buffer and send a NAK
;		this will instruct XMODEM to repeat the transmission of the
;		buffer
;
BadCrc:		
		rcall	Flush			; flush the input port
		ldi		char, NAK			;
		rcall	Put_Chr			; send NAK to resend block
		rjmp	StartBlk		; start over, get the block again			
;
;		Successfully received a record with correct CRC. We expect
;		files which start with a two byte load address, as is the
;		default for binary files created by TASS and 64tass. So
;		we need to put the first two bytes of the first record
;		as start address. A flag (bflag) tracks if we already
;		have stored the starting address
;
GoodCrc:
		ldi		yl, 0x02		; Start of Data Record
CopyRec:	
		lds		temp, bflag		; do we have an address
		tst		temp			; do we have an address
		brne	HaveAddr		; yes we have an address
		rcall	imlprep
		sts		bflag, one		; Remember we have an address
HaveAddr:
		ldi		temp, LDA_
		rcall	imlbyte
		ld		temp, Y+
		rcall	imlbyte
		ldi		temp, STA_
		rcall	imlbyte
		mov		temp, r24
		rcall	imlbyte
		mov		temp, r25
		rcall	imlbyte
		ldi		temp, 0xff
		rcall	imlbyte
		adiw	r24:r25, 1
		ldi		temp, JMP_
		rcall	imlbyte
		ldi		temp, 0x00
		rcall	imlbyte
		ldi		temp, 0x03
		rcall	imlbyte
		cpi		yl, 0x82		; all 128 bytes done?	
		brne	CopyRec			; still more bytes to go
;
;
;
IncBlk:							; done
		lds		temp, blkno
		inc		temp			; Inc the block #
		sts		blkno, temp
		ldi		char, ACK		; send ACK
		call	Put_Chr			;
		rjmp	StartBlk		; get next block
;
;		EOT - end of text received send final ACK and finish
;
Done:		
		ldi		char, ACK		; last block, send ACK and exit.
		rcall	Put_Chr			;
		rcall	Flush			; get leftover characters, if any
		rcall	Print_Good		;
		call	videostart
		rcall	imlstop
		sbi		RESETPIN
		cbi		IMLPIN
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2
		sbi		PORTD, PHI2
		sts		a1l, zero
		rcall	imlclock
		cbi		RESETPIN		

xexit:
		pop		yh
		pop		yl
		pop		xh
		pop		xl 	 
		pop		r25
		pop		r24

		ret						;


;--------------------------------------------------------------------------
;
;		Now we need to reset the processor and start down
;		loading
;
;--------------------------------------------------------------------------
imlprep:
		rcall	imlstop			; stop the clock
		sbi		RESETPIN
		sbi		IMLPIN
		sbi		PORTD, PHI2		; Note that consecutive sbi/cbi
		cbi		PORTD, PHI2		; create a clock requency of 4 AVR
		sbi		PORTD, PHI2		; cycles in our case 5.5MHz.
		cbi		PORTD, PHI2
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2
		sbi		PORTD, PHI2
		cbi		RESETPIN
		cbi		PORTD, PHI2		; this for some reason does not

		sbi		PORTD, PHI2
		cbi		PORTD, PHI2		; This detects that reset is de-asserted
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2		; Cylce 1
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2		; Cylce 2
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2		; Cylce 3
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2		; Cylce 4
		sbi		PORTD, PHI2
		cbi		PORTD, PHI2		; Cylce 5
		
		ld		r24, Y+
		ld		r25, Y+			; Get pointer
		
		ldi		temp, 0x00
		rcall	imlbyte
		ldi		temp, 0x03
		rcall	imlbyte
		ldi		temp, STA_
		rcall	imlbyte
		ldi		temp, 0x06			; Assume we clear CXROM
		brtc	clrcx				; Yes 
		ldi		temp, 0x07			; Else we set CXROM
clrcx:	rcall	imlbyte
		ldi		temp, 0xc0			; $C006/7
		rcall	imlbyte
		ldi		temp, 0xff			; The write cycle
		rcall	imlbyte
		ret

;--------------------------------------------------------------------------
;
; Write 256 bytes to Flash
;
;	Input	RamBuff contains 256bytes / 128words of data to write to flash
;			ZSAV (rzsav:zsavh:zsav) contains the byte address in flash
;
;	Output	Flash updated
;			ZSAV incremented
;			X set to start of RamBuff
;--------------------------------------------------------------------------
WritePage:
		in		temp, SREG			; save sreg
		push	temp
		cli							; disable irq's
		lds		zl, zptr
		lds		zh, zptr+1
		ldi		xl, low(RamBuff)
		ldi		xh, high(RamBuff)	; start of buffer
fploop:
		ld		r0, X+
		ld		r1, X+
stbf1:
		in		char, SPMCSR		; get status
		sbrc	char, SPMEN			; wait for spm to complete
		rjmp	stbf1
		ldi		char, (1<<SPMEN)	; Write word to buffer
		out		SPMCSR, char		; enable SPM only - write to int buffer
		spm
		subi	zl, 0xfe
		brne	fploop
erase1:
		in		char, SPMCSR		; get status
		sbrc	char, SPMEN			; skip if spm complete
		rjmp	erase1				; wait for spm to complete
		ldi		char, (1<<PGERS) | (1<<SPMEN)
		out		SPMCSR, char		; PGERS+SELFPRGEN (erase page)
		spm
write2:
		in		char, SPMCSR		; get status
		sbrc	char, SPMEN			; skip if spm complete
		rjmp	write2				; wait for spm to complete
		ldi		char, (1<<PGWRT) | (1<<SPMEN)
		out		SPMCSR, char		; PGWRT+SELFPRGEN(write page)	
		spm
		inc		zh					; go to next page
write3:
		in		char, SPMCSR		; get status
		sbrc	char, SPMEN			; skip if spm complete
		rjmp	write3				; wait for spm to complete
		ldi		char, (1<<RWWSRE) | (1<<SPMEN)
		out		SPMCSR, char		; RWWSRE+SELFPRGEN (Enable RWW reads)
		spm
		sts		zptr, zl
		sts		zptr+1, zh
		ldi		xl, low(RamBuff)
		ldi		xh, high(RamBuff)	; start of buffer
		pop		temp
		out		SREG, temp			; restore sreg & int state
		ret
		
;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;
; subroutines
;
;								;
;
Flush:							; flush receive buffer
		ldi		char, 1
Flush1:	
		rcall	GetByte			; read the port
		brcs	Flush			; if chr recvd, wait for another
		ret						; else done
;
PrintMsg:
		ldi		zh, high(Msg*2)	; PRINT starting message
		ldi		zl, low(Msg*2)	;
PrtMsg1:
		lpm   	char, Z+		
		tst		char
		breq	PrtMsg2			
		rcall	Put_Chr
		rjmp	PrtMsg1
PrtMsg2:
		ret

Msg:
		.db		"Begin XMODEM/CRC transfer. Press <Esc> to abort..."
		.db  	CR, LF, 0x00, 0x00
;
Print_Err:
		ldi		zh, high(ErrMsg*2) ; PRINT Error message
		ldi		zl, low(ErrMsg*2)  ;
PrtErr1:
		lpm   	char, Z+
		tst		char
		breq	PrtErr2
		rcall	Put_Chr
		rjmp	PrtErr1 
PrtErr2:
		ret

ErrMsg:
		.db 	"Upload Error! "
		.db  	CR, LF, 0x00, 0x00
;
Print_Good:
		ldi		zh, high(GoodMsg*2) ; PRINT Good Transfer message
		ldi		zl, low(GoodMsg*2)  ;
Prtgood1:
		lpm   	char, Z+
		tst		char
		breq	Prtgood2
		rcall	Put_Chr
		rjmp	Prtgood1
Prtgood2:
		ret

GoodMsg:
		.db 	"Upload Successful!"
		.db 	CR, LF, 0x00, 0x00
;======================================================================
;  I/O Device Specific Routines
;
;  Two routines are used to communicate with the I/O device.
;
; "Get_Chr" routine will scan the input port for a character with a 
; time-out. We use Timer 3 to create a 1 second interval. 
;
;	input	char	time-out in seconds
;	output	carry	is set if we got a character, cleared if time-out
;			char	contains the character if any, else zero
;
;	Prepare a 1 second time-out, we use timer 1 and use a prescaler
;	of 1024. At 25MHz clock this gives a counter clock of ~25kHz.
;	We set the TCNT1 register to -F_CPU/1024 and let the counter count
;	normally, this will update the overflow flag when the count
;	reaches 0xFFFF, which is approximatively within one second
;
GetByte:						; wait for chr input and cycle timing loop
	push	char
	ldi		char, high(-F_CPU/1024)
	sts		TCNT1H, char
	ldi		char, low(-F_CPU/1024)
	sts		TCNT1L, char		; 
	ldi		char, (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10)
	sts		TCCR1A, zero
	ldi		char, (0 << WGM13) | (0 << WGM12) | (1<<CS12) | (0<<CS11) | (1<<CS10)
	sts		TCCR1B, char		; Start Timer

GetByte1:
	lds		char, UCSR0A
	sbrc	char, RXC0			; 
	rjmp	GetByte2			; got a character
	
	sbis	TIFR1, TOV1
	rjmp	GetByte1			; wait
	sbi		TIFR1, TOV1			; reset overflow
	sts		TCCR1B, zero		; stop timer
	pop		char
	dec		char
	brne	GetByte

	sbi		TIFR1, TOV1			; reset overflow
	sts		TCCR1B, zero		; stop timer
	clc
	ret
GetByte2:
	pop		char
	lds		char, UDR0
	sts		TCCR1B, zero
	sbi		TIFR1, TOV1
	sec
	ret
;
; "Put_Chr" routine will write one byte to the output port. It's alright
; if this routine waits for the port to be ready.
;
;	input		char		contains the character to be sent via USART
;	output					nothing
;
Put_Chr:
		push	char
Put_Chr2:
	   	lds 	char, UCSR0A	; 
		sbrs	char, UDRE0
       	rjmp	Put_chr2		; is tx buffer empty
       	pop		char
       	sts		UDR0, char		; put character to Port
       	ret						; done
;=========================================================================
;
;
;  CRC subroutines 
;
;
UpdCrc:
		lds		temp, crc+1
		eor		char, temp
		ldi		zl, low(crchi*2)
		ldi		zh, high(crchi*2)
		add		zl, char
		adc		zh, zero
		lpm		char, Z
		lds		temp, crc
		eor		char, temp
		sts		crc+1, char
		dec		zh
		lpm		char, Z
		sts		crc, char
		ret  

; low byte CRC lookup table

crclo:
 .db 0x00,0x21,0x42,0x63,0x84,0xA5,0xC6,0xE7,0x08,0x29,0x4A,0x6B,0x8C,0xAD,0xCE,0xEF
 .db 0x31,0x10,0x73,0x52,0xB5,0x94,0xF7,0xD6,0x39,0x18,0x7B,0x5A,0xBD,0x9C,0xFF,0xDE
 .db 0x62,0x43,0x20,0x01,0xE6,0xC7,0xA4,0x85,0x6A,0x4B,0x28,0x09,0xEE,0xCF,0xAC,0x8D
 .db 0x53,0x72,0x11,0x30,0xD7,0xF6,0x95,0xB4,0x5B,0x7A,0x19,0x38,0xDF,0xFE,0x9D,0xBC
 .db 0xC4,0xE5,0x86,0xA7,0x40,0x61,0x02,0x23,0xCC,0xED,0x8E,0xAF,0x48,0x69,0x0A,0x2B
 .db 0xF5,0xD4,0xB7,0x96,0x71,0x50,0x33,0x12,0xFD,0xDC,0xBF,0x9E,0x79,0x58,0x3B,0x1A
 .db 0xA6,0x87,0xE4,0xC5,0x22,0x03,0x60,0x41,0xAE,0x8F,0xEC,0xCD,0x2A,0x0B,0x68,0x49
 .db 0x97,0xB6,0xD5,0xF4,0x13,0x32,0x51,0x70,0x9F,0xBE,0xDD,0xFC,0x1B,0x3A,0x59,0x78
 .db 0x88,0xA9,0xCA,0xEB,0x0C,0x2D,0x4E,0x6F,0x80,0xA1,0xC2,0xE3,0x04,0x25,0x46,0x67
 .db 0xB9,0x98,0xFB,0xDA,0x3D,0x1C,0x7F,0x5E,0xB1,0x90,0xF3,0xD2,0x35,0x14,0x77,0x56
 .db 0xEA,0xCB,0xA8,0x89,0x6E,0x4F,0x2C,0x0D,0xE2,0xC3,0xA0,0x81,0x66,0x47,0x24,0x05
 .db 0xDB,0xFA,0x99,0xB8,0x5F,0x7E,0x1D,0x3C,0xD3,0xF2,0x91,0xB0,0x57,0x76,0x15,0x34
 .db 0x4C,0x6D,0x0E,0x2F,0xC8,0xE9,0x8A,0xAB,0x44,0x65,0x06,0x27,0xC0,0xE1,0x82,0xA3
 .db 0x7D,0x5C,0x3F,0x1E,0xF9,0xD8,0xBB,0x9A,0x75,0x54,0x37,0x16,0xF1,0xD0,0xB3,0x92
 .db 0x2E,0x0F,0x6C,0x4D,0xAA,0x8B,0xE8,0xC9,0x26,0x07,0x64,0x45,0xA2,0x83,0xE0,0xC1
 .db 0x1F,0x3E,0x5D,0x7C,0x9B,0xBA,0xD9,0xF8,0x17,0x36,0x55,0x74,0x93,0xB2,0xD1,0xF0 

; hi byte CRC lookup table

crchi:
 .db 0x00,0x10,0x20,0x30,0x40,0x50,0x60,0x70,0x81,0x91,0xA1,0xB1,0xC1,0xD1,0xE1,0xF1
 .db 0x12,0x02,0x32,0x22,0x52,0x42,0x72,0x62,0x93,0x83,0xB3,0xA3,0xD3,0xC3,0xF3,0xE3
 .db 0x24,0x34,0x04,0x14,0x64,0x74,0x44,0x54,0xA5,0xB5,0x85,0x95,0xE5,0xF5,0xC5,0xD5
 .db 0x36,0x26,0x16,0x06,0x76,0x66,0x56,0x46,0xB7,0xA7,0x97,0x87,0xF7,0xE7,0xD7,0xC7
 .db 0x48,0x58,0x68,0x78,0x08,0x18,0x28,0x38,0xC9,0xD9,0xE9,0xF9,0x89,0x99,0xA9,0xB9
 .db 0x5A,0x4A,0x7A,0x6A,0x1A,0x0A,0x3A,0x2A,0xDB,0xCB,0xFB,0xEB,0x9B,0x8B,0xBB,0xAB
 .db 0x6C,0x7C,0x4C,0x5C,0x2C,0x3C,0x0C,0x1C,0xED,0xFD,0xCD,0xDD,0xAD,0xBD,0x8D,0x9D
 .db 0x7E,0x6E,0x5E,0x4E,0x3E,0x2E,0x1E,0x0E,0xFF,0xEF,0xDF,0xCF,0xBF,0xAF,0x9F,0x8F
 .db 0x91,0x81,0xB1,0xA1,0xD1,0xC1,0xF1,0xE1,0x10,0x00,0x30,0x20,0x50,0x40,0x70,0x60
 .db 0x83,0x93,0xA3,0xB3,0xC3,0xD3,0xE3,0xF3,0x02,0x12,0x22,0x32,0x42,0x52,0x62,0x72
 .db 0xB5,0xA5,0x95,0x85,0xF5,0xE5,0xD5,0xC5,0x34,0x24,0x14,0x04,0x74,0x64,0x54,0x44
 .db 0xA7,0xB7,0x87,0x97,0xE7,0xF7,0xC7,0xD7,0x26,0x36,0x06,0x16,0x66,0x76,0x46,0x56
 .db 0xD9,0xC9,0xF9,0xE9,0x99,0x89,0xB9,0xA9,0x58,0x48,0x78,0x68,0x18,0x08,0x38,0x28
 .db 0xCB,0xDB,0xEB,0xFB,0x8B,0x9B,0xAB,0xBB,0x4A,0x5A,0x6A,0x7A,0x0A,0x1A,0x2A,0x3A
 .db 0xFD,0xED,0xDD,0xCD,0xBD,0xAD,0x9D,0x8D,0x7C,0x6C,0x5C,0x4C,0x3C,0x2C,0x1C,0x0C
 .db 0xEF,0xFF,0xCF,0xDF,0xAF,0xBF,0x8F,0x9F,0x6E,0x7E,0x4E,0x5E,0x2E,0x3E,0x0E,0x1E 
;
;
; End of File
;
